import gc
import random
import numpy as np
import os
import time
from datetime import datetime
from collections import deque
from sklearn import metrics
from scipy.sparse import csr_matrix
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch import optim
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool, GlobalAttention, Set2Set
import pickle
import copy
import argparse
import networkx as nx
import scipy.sparse as sp
from torch_geometric import loader
import en_core_web_lg
import pandas as pd
import networkx as nx
from torch_geometric.data import Data
from torch_geometric import loader
from torch_geometric.nn import GCNConv
from torch.distributions import Categorical, MultivariateNormal
import sys
from sklearn.metrics import silhouette_score, calinski_harabasz_score
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader


def currentTime():
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S')


class args_define:
    def __init__(self, **kwargs):
        # Paths
        self.file_path = kwargs.get('file_path', '../model/model_saved/hcrc/')
        self.result_path = kwargs.get('result_path', '../model/model_saved/hcrc/res.txt')

        # Task
        self.task = kwargs.get('task', 'DRL')

        # Layers
        self.layers = kwargs.get('layers', '[256]')

        # Hidden units
        self.N_pred_hid = kwargs.get('N_pred_hid', 64)
        self.G_pred_hid = kwargs.get('G_pred_hid', 16)

        # Evaluation frequency
        self.eval_freq = kwargs.get('eval_freq', 5)

        # Moving Average Decay
        self.mad = kwargs.get('mad', 0.9)

        # Learning rates
        self.Glr = kwargs.get('Glr', 0.0000006)
        self.Nlr = kwargs.get('Nlr', 0.00001)

        # Early Stopping Criteria
        self.Ges = kwargs.get('Ges', 50)
        self.Nes = kwargs.get('Nes', 2000)

        # Device
        self.device = kwargs.get('device', 0)

        # Epochs
        self.Gepochs = kwargs.get('Gepochs', 105)
        self.Nepochs = kwargs.get('Nepochs', 100)

        # Store all arguments in a single attribute
        self.args = argparse.Namespace(**{
            'file_path': self.file_path,
            'result_path': self.result_path,
            'task': self.task,
            'layers': self.layers,
            'N_pred_hid': self.N_pred_hid,
            'G_pred_hid': self.G_pred_hid,
            'eval_freq': self.eval_freq,
            'mad': self.mad,
            'Glr': self.Glr,
            'Nlr': self.Nlr,
            'Ges': self.Ges,
            'Nes': self.Nes,
            'device': self.device,
            'Gepochs': self.Gepochs,
            'Nepochs': self.Nepochs
        })

class HCRC:
    def __init__(self, args, dataset):
        self.args = args
        self.dataset = dataset

    def detection(self):
        for i in range(22):
            print("************Message Block " + str(i) + " start! ************")
            # Node-level learning
            embedder_N = Node_ModelTrainer(args, i)
            Node_emb, label = embedder_N.get_embedding()
            # Graph-level learning
            embedder_G = Graph_ModelTrainer(args, i)
            Graph_emb, label = embedder_G.get_embedding()
            # combining vectors
            if i == 0:
                all_embeddings = np.concatenate((Graph_emb, Node_emb), axis=1)
                all_label = label
            else:
                temp = np.concatenate((Graph_emb, Node_emb), axis=1)
                all_embeddings = np.concatenate((all_embeddings, temp), axis=0)
                all_label = all_label + label
            all_embeddings = torch.tensor(all_embeddings)
            all_embeddings = F.normalize(all_embeddings, dim=-1, p=2).detach().cpu().numpy()

            if i == 0:
                pred_y = evaluate_fun(all_embeddings, label, i, None, args.result_path, args.task)
                all_pred_y = pred_y
            else:
                pred_y = evaluate_fun(all_embeddings, label, i, all_pred_y, args.result_path, args.task)
                all_pred_y = all_pred_y + pred_y
            print("************Message Block " + str(i) + " end! ************\n\n")

        predictions = all_pred_y
        ground_truths = all_label

        return predictions, ground_truths

    def evaluate(self, predictions, ground_truths):
        print("************Evaluation start! ************")
        ars = metrics.adjusted_rand_score(ground_truths, predictions)

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)

        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)

        print(f"Model Adjusted Rand Index (ARI): {ars}")
        print(f"Model Adjusted Mutual Information (AMI): {ami}")
        print(f"Model Normalized Mutual Information (NMI): {nmi}")
        return ars, ami, nmi


class EMA:  # Exponential Moving Average
    def __init__(self, beta, epochs):
        super().__init__()
        self.beta = beta
        self.step = 0
        self.total_steps = epochs

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new


def get_task(strs):
    tasks = ["DRL", "random", "semi-supervised", "traditional"]
    if len(strs) == 1:
        return "DRL"
    if ("--task" in strs) and len(strs) == 2:
        return "DRL"
    if ("--task" not in strs) or len(strs) != 3:
        return False
    elif strs[-1] not in tasks:
        return False
    else:
        return strs[-1]


def init_weights(m):  # Model parameter initialization
    if type(m) == nn.Linear:
        torch.nn.init.xavier_uniform_(m.weight)
        m.bias.data.fill_(0.01)


def sim(z1, z2):
    z1 = F.normalize(z1)
    z2 = F.normalize(z2)
    return torch.mm(z1, z2.t())


def semi_loss(z1, z2):
    f = lambda x: torch.exp(x / 0.05)
    refl_sim = f(sim(z1, z1))
    between_sim = f(sim(z1, z2))

    return -torch.log(between_sim.diag() / (refl_sim.sum(1) + between_sim.sum(1) - refl_sim.diag()))


def get_loss(h1, h2):
    l1 = semi_loss(h1, h2)
    l2 = semi_loss(h2, h1)

    ret = (l1 + l2) * 0.5
    ret = ret.mean()

    return ret


def update_moving_average(ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = ema_updater.update_average(old_weight, up_weight)


def set_requires_grad(model, val):
    # set require_grad
    for p in model.parameters():
        p.requires_grad = val


def enumerateConfig(args):
    args_names = []
    args_vals = []
    for arg in vars(args):
        args_names.append(arg)
        args_vals.append(getattr(args, arg))

    return args_names, args_vals


def config2string(args):
    args_names, args_vals = enumerateConfig(args)
    st = ''
    for name, val in zip(args_names, args_vals):
        if val == False:
            continue
        if name not in ['device', 'root', 'epochs', 'isAnneal', 'dropout', 'warmup_step', 'clus_num_iters']:
            st_ = "{}_{}_".format(name, val)
            st += st_

    return st[:-1]


def printConfig(args):
    args_names, args_vals = enumerateConfig(args)
    print(args_names)
    print(args_vals)


class embedder:
    def __init__(self, args):
        self.args = args
        self.hidden_layers = eval(args.layers)


class Encoder(nn.Module):

    def __init__(self, layer_config):
        super().__init__()
        self.stacked_gnn = nn.ModuleList(
            [GCNConv(layer_config[i - 1], layer_config[i]) for i in range(1, len(layer_config))])
        self.stacked_bns = nn.ModuleList(
            [nn.BatchNorm1d(layer_config[i], momentum=0.01) for i in range(1, len(layer_config))])
        self.stacked_prelus = nn.ModuleList([nn.PReLU() for _ in range(1, len(layer_config))])

    def forward(self, x, edge_index):
        for i, gnn in enumerate(self.stacked_gnn):
            x = gnn(x, edge_index, edge_weight=None)
            x = self.stacked_bns[i](x)
            x = self.stacked_prelus[i](x)

        return x


M = [20254, 28976, 30467, 32302, 34312, 36146, 37422, 42700, 44260, 45623, 46719,
     47951, 51188, 53160, 56116, 58665, 59575, 62251, 64138, 65537, 66430, 68840]


def DRL_cluster(all_embeddings, block_num, pred_label):
    para = 0.1
    if block_num == 0:
        print("Evaluating initial message block...")
        start_time = time.time()
        sp = SinglePass(0.87, all_embeddings, 0, pred_label, M[0], None, para, 0, sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
    else:
        print("Using DRL-Single-Pass to learn threshold...")
        global_step = 0
        agent = PPO([5], 1, continuous=True)
        sp_sim = SinglePass(0.6, all_embeddings, 1, pred_label, M[block_num] - M[block_num - 1], agent, para,
                            M[block_num - 1] - 2000, sim=True)

        global_step = sp_sim.global_step
        sp = SinglePass(0.6, all_embeddings, 1, pred_label, M[block_num] - M[block_num - 1], agent, para,
                        M[block_num - 1] - 2000, sim=False)

    return sp.cluster_result, sp.sim_threshold


def random_cluster(all_embeddings, block_num, pred_label):
    threshold = random.uniform(0.6, 0.8)
    if block_num == 0:
        print("Evaluating initial message block...")
        start_time = time.time()
        sp = SinglePass(0.87, all_embeddings, 0, pred_label, M[0], None, 0, 0, sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
        threshold = 0.87
    else:
        print("Evaluating message block...")
        start_time = time.time()
        sp = SinglePass(threshold, all_embeddings, 2, pred_label, M[block_num] - M[block_num - 1], None, 0, 0,
                        sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
    return sp.cluster_result, threshold


def semi_cluster(all_embeddings, label, block_num, pred_label):
    if block_num == 0:
        print("Evaluating initial message block...")
        start_time = time.time()
        sp = SinglePass(0.87, all_embeddings, 0, pred_label, M[0], None, 0, 0, sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
        threshold = 0.87
    else:
        print("Evaluating message block...")
        start_time = time.time()
        embeddings = all_embeddings.tolist()
        size = M[block_num] - M[block_num - 1]
        embeddings = embeddings[0:len(embeddings) - int(size * 0.9)]
        pre_label = pred_label[0:len(embeddings)]

        size = len(embeddings) - M[block_num - 1]
        embeddings = np.array(embeddings)
        thresholds = [0.6, 0.65, 0.7, 0.75, 0.8]
        s1s = []
        for t in thresholds:
            sp = SinglePass(t, embeddings, 2, pre_label, size, None, 0, 0, sim=False)
            true_label = label[0:len(sp.cluster_result)]
            s1 = metrics.normalized_mutual_info_score(true_label, sp.cluster_result, average_method='arithmetic')
            s1s.append(s1)
        index = s1s.index(max(s1s))
        sp = SinglePass(thresholds[index], all_embeddings, 2, pred_label, M[block_num] - M[block_num - 1], None, 0, 0,
                        sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
        threshold = thresholds[index]
    return sp.cluster_result, threshold


def NMI_cluster(all_embeddings, label, block_num, pred_label):
    if block_num == 0:
        print("Evaluating initial message block...")
        start_time = time.time()
        sp = SinglePass(0.87, all_embeddings, 0, pred_label, M[0], None, 0, 0, sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
        threshold = 0.87
    else:
        print("Evaluating message block...")
        start_time = time.time()
        thresholds = [0.6, 0.65, 0.7, 0.75, 0.8]
        s1s = []
        for t in thresholds:
            sp = SinglePass(t, all_embeddings, 2, pred_label, M[block_num] - M[block_num - 1], None, 0, 0, sim=False)
            s1 = metrics.normalized_mutual_info_score(label, sp.cluster_result, average_method='arithmetic')
            s1s.append(s1)
        index = s1s.index(max(s1s))
        sp = SinglePass(thresholds[index], all_embeddings, 2, pred_label, M[block_num] - M[block_num - 1], None, 0, 0,
                        sim=False)
        end_time = time.time()
        run_time = end_time - start_time
        print("Done! " + "It takes " + str(int(run_time)) + " seconds.\n")
        threshold = thresholds[index]
    return sp.cluster_result, threshold


def evaluate_fun(all_embeddings, label, block_num, pred_label, result_path, task):
    if task == "DRL":
        y_pred, threshold = DRL_cluster(all_embeddings, block_num, pred_label)
    elif task == "random":
        y_pred, threshold = random_cluster(all_embeddings, block_num, pred_label)
    elif task == "semi-supervised":
        y_pred, threshold = semi_cluster(all_embeddings, label, block_num, pred_label)
    elif task == "traditional":
        y_pred, threshold = NMI_cluster(all_embeddings, label, block_num, pred_label)

    # NMI
    s1 = metrics.normalized_mutual_info_score(label, y_pred, average_method='arithmetic')
    # AMI
    s2 = metrics.adjusted_mutual_info_score(label, y_pred, average_method='arithmetic')
    # ARI
    s3 = metrics.adjusted_rand_score(label, y_pred)

    print('** Theta:{:.2f} **\n'.format(threshold))
    print('** NMI: {:.2f} **\n'.format(s1))
    print('** AMI: {:.2f} **\n'.format(s2))
    print('** ARI: {:.2f} **\n'.format(s3))
    result = '\nmessage_block_' + str(block_num) + '\nthreshold: {:.2f} '.format(
        threshold) + '\n** NMI: {:.2f} **\n'.format(s1) + '** AMI: {:.2f} **\n'.format(
        s2) + '** ARI: {:.2f} **\n'.format(s3)

    if not os.path.exists(result_path):
        pass
    else:
        with open(result_path, encoding='utf-8') as file:
            content = file.read()
        result = content.rstrip() + result
    file = open(result_path, mode='w')
    file.write(result)
    file.close()
    return y_pred


class SinglePass:
    def __init__(self, sim_threshold, data, flag, label, size, agent, para, sim_init, sim=False, global_step=0):
        self.device = torch.device('cuda:0')
        self.text_vec = None  #
        self.topic_serial = None
        self.topic_cnt = 0
        self.sim_threshold = sim_threshold

        self.done_data = data[0:data.shape[0] - size]
        self.new_data = data[data.shape[0] - size:]
        self.done_label = label

        if flag == 0 or flag == 2:
            self.cluster_result = self.run_cluster(flag, size)
        else:
            self.agent = agent
            self.scheme = ["state", "action", "reward", "done", "log_prob"]
            self.global_step = global_step
            self.sim = sim
            if self.sim:
                start_time = time.time()
                self.cluster_result = self.run_cluster_sim(flag, size, para, sim_init, sim, data)
                end_time = time.time()
                self.time = end_time - start_time
                print("Creating Environment Done! " + "It takes " + str(int(self.time)) + " seconds.")
            else:
                start_time = time.time()
                self.pseudo_labels = self.run_cluster_init(0.6, size)
                if flag == 1:
                    self.text_vec = self.done_data
                    self.topic_serial = copy.deepcopy(self.done_label)
                    self.topic_cnt = max(self.topic_serial)
                state = self.get_state(sim, sim_init, data)
                action, action_log_prob = self.agent.select_action(state)
                # action projection
                sim_threshold = torch.clamp(action, -1, 1).detach()
                sim_threshold += 7
                sim_threshold /= 10
                self.sim_threshold = sim_threshold.item()
                end_time = time.time()
                self.time = end_time - start_time
                print("Getting Threshold Done! " + "It takes " + str(int(self.time)) + " seconds. ")
                print("Threshold is " + str(self.sim_threshold) + ".\n")
                print("Evaluating message block...")
                start_time = time.time()
                self.cluster_result = self.run_cluster(flag, size)  # clustering
                end_time = time.time()
                self.time = end_time - start_time
                print("Done! " + "It takes " + str(int(self.time)) + " seconds.\n")

    def clustering(self, sen_vec):
        if self.topic_cnt == 0:
            self.text_vec = sen_vec
            self.topic_cnt += 1
            self.topic_serial = [self.topic_cnt]
        else:
            sim_vec = np.dot(sen_vec, self.text_vec.T)
            max_value = np.max(sim_vec)

            topic_ser = self.topic_serial[np.argmax(sim_vec)]
            self.text_vec = np.vstack([self.text_vec, sen_vec])

            if max_value >= self.sim_threshold:
                self.topic_serial.append(topic_ser)
            else:
                self.topic_cnt += 1
                self.topic_serial.append(self.topic_cnt)

    def clustering_init(self, t, sen_vec):
        if self.topic_cnt_init == 0:
            self.text_vec_init = sen_vec
            self.topic_cnt_init += 1
            self.topic_serial_init = [self.topic_cnt_init]
        else:
            sim_vec = np.dot(sen_vec, self.text_vec_init.T)
            max_value = np.max(sim_vec)

            topic_ser = self.topic_serial_init[np.argmax(sim_vec)]
            self.text_vec_init = np.vstack([self.text_vec_init, sen_vec])

            if max_value >= t:
                self.topic_serial_init.append(topic_ser)
            else:
                self.topic_cnt_init += 1
                self.topic_serial_init.append(self.topic_cnt_init)

    def run_cluster_init(self, t, size):
        self.text_vec_init = []
        self.topic_serial_init = []
        self.topic_cnt_init = 0
        for vec in self.new_data:
            self.clustering_init(t, vec)
        return self.topic_serial_init

    def run_cluster_sim(self, flag, size, para, sim_init, sim, data):
        self.text_vec = []
        self.topic_serial = []
        self.topic_cnt = 0
        if flag == 1:
            self.text_vec = self.done_data
            self.topic_serial = copy.deepcopy(self.done_label)
            self.topic_cnt = max(self.topic_serial)
        for i, vec in enumerate(self.new_data):
            self.global_step += 1
            if i > 200:
                break
            if i > self.new_data.shape[0] * para:
                break
            state = self.get_state(sim, sim_init, data)

            action, action_log_prob = self.agent.select_action(state)
            self.sim_threshold = action.item()
            self.clustering(vec)

            reward = self.get_reward(sim_init, data)
            done = False
            transition = make_transition(self.scheme, state, action, reward, done, action_log_prob)
            self.agent.add_buffer(transition)
            if self.global_step % 200 == 0:
                self.agent.learn()

        return self.topic_serial[len(self.topic_serial) - size:]

    def run_cluster(self, flag, size):
        self.text_vec = []
        self.topic_serial = []
        self.topic_cnt = 0
        if flag == 1 or flag == 2:
            self.text_vec = self.done_data
            self.topic_serial = copy.deepcopy(self.done_label)
            self.topic_cnt = max(self.topic_serial)
        for i, vec in enumerate(self.new_data):
            self.clustering(vec)
        return self.topic_serial[len(self.topic_serial) - size:]

    def get_center(self, label, data):
        centers = []
        indexs_per_cluster = []
        label_u = list(set(label))
        for i in range(len(label_u)):
            indexs = [False] * data.shape[0]
            tmp_indexs_text = []
            for j in range(len(indexs)):
                if label[j] == label_u[i]:
                    indexs[j] = True
                    tmp_indexs_text.append(j)
            center = np.mean(data[indexs], 0).tolist()
            centers.append(center)
            indexs_per_cluster.append(tmp_indexs_text)
        return centers, indexs_per_cluster

    def get_info_cluster(self, text_vec, indexs_per_cluster):  # Get detailed clustering results
        res = []
        for i in range(len(indexs_per_cluster)):
            tmp_vec = []
            for j in range(len(indexs_per_cluster[i])):
                tmp_vec.append(text_vec[indexs_per_cluster[i][j]])
            tmp_vec = np.array(tmp_vec)
            res.append(tmp_vec)
        return res

    def get_state(self, sim, sim_init, data):  # get state of RL
        state = []
        if sim:
            data = data[sim_init:len(self.topic_serial)]
            topic_serial = self.topic_serial[sim_init:]
        else:
            data = self.new_data
            topic_serial = self.pseudo_labels
        centers, indexs_per_cluster = self.get_center(topic_serial, data)

        centers = np.array(centers)
        neighbor_dists = np.dot(centers, centers.T)

        neighbor_dists = np.nan_to_num(neighbor_dists, 0.0001)
        # the minimum neighbor distance
        state.append(neighbor_dists.min())
        # the average separation distance
        state.append((neighbor_dists.mean() * max(topic_serial) - 1) / max(topic_serial))
        info_of_cluster = self.get_info_cluster(data, indexs_per_cluster)

        coh_dists = 0
        for cluster in info_of_cluster:
            if cluster.shape[0] == 1:
                continue
            else:
                sums = cluster.shape[0] * (cluster.shape[0] - 1) / 2
            tmp_vec = np.array(cluster)
            cohdist = np.dot(tmp_vec, tmp_vec.T)
            if cohdist.max() > coh_dists:
                coh_dists = cohdist.max()
        #         Dunn index
        state.append(neighbor_dists.min() / coh_dists)

        # Sum of intra-group error squares
        SSE = 0
        SSEE = 0
        for i in range(len(indexs_per_cluster)):
            sumtmp = 0
            for j in range(len(indexs_per_cluster[i])):
                tmp = np.dot(data[indexs_per_cluster[i][j]].T, centers[i])
                SSE = SSE + (tmp) ** 2
                sumtmp = sumtmp + (tmp) ** 2
            SSEE = SSEE + sumtmp / len(indexs_per_cluster[i])
        #         state.append(SSE)

        # Sum of squared errors between groups
        SSR = 0
        SSRR = 0
        for i in range(len(centers)):
            SSR = SSR + np.dot(centers[i].T, centers.mean(axis=0))
            SSRR = SSRR + np.dot(centers[i].T, centers.mean(axis=0)) ** 2
        SSRR = SSRR / max(topic_serial)
        #         state.append(SSR)
        # the average cohesion distance
        coh_dists = 0
        for cluster in info_of_cluster:
            if cluster.shape[0] == 1:
                continue
            else:
                sums = cluster.shape[0] * (cluster.shape[0] - 1) / 2
            tmp_vec = np.array(cluster)
            cohdist = np.dot(tmp_vec, tmp_vec.T)
            cohdist = np.maximum(cohdist, -cohdist)
            coh_dists = coh_dists + (cohdist.sum() - cluster.shape[0]) / (2 * sums + 0.0001)
        state.append(coh_dists / max(topic_serial))

        state.append(silhouette_score(data, topic_serial, metric='euclidean'))
        return np.array(state)

    def get_reward(self, sim_init, data):  # get reward of RL

        data = data[sim_init:len(self.topic_serial)]
        topic_serial = self.topic_serial[sim_init:]

        return calinski_harabasz_score(data, topic_serial)


class Node_ModelTrainer(embedder):

    def __init__(self, args, block_num):
        embedder.__init__(self, args)
        self._args = args
        self.block_num = block_num
        self._init()

    def _init(self):
        block_num = self.block_num
        args = self._args
        os.environ["CUDA_VISIBLE_DEVICES"] = str(args.device)
        self._device = f'cuda:{args.device}' if torch.cuda.is_available() else "cpu"
        torch.cuda.set_device(self._device)

        if block_num == 0 or block_num == 1:
            args.Nes = 200
        else:
            args.Nes = 2000
        self.true_label = []
        # load data
        self._loader, self.true_label = get_Node_Dataset(block_num)

        layers = [302] + self.hidden_layers
        self._model = NodeLevel(layers, args)

        self._optimizer = optim.AdamW(params=self._model.parameters(), lr=args.Nlr, weight_decay=1e-5)
        self.train()
        self.all_embeddings = F.normalize(self.all_embeddings, dim=-1, p=2).detach().cpu().numpy()

    def get_embedding(self):
        return self.all_embeddings, self.true_label

    def train(self):

        loss_t = 1e10
        cnt_wait = 0
        # Start Model Training
        print("----Node-Level Training Start! ----\n")
        for epoch in range(self._args.Nepochs):
            losses = []
            embs = []
            for batch in self._loader:
                emb, loss = self._model(batch)
                self._optimizer.zero_grad()
                loss.backward()
                self._optimizer.step()
                self._model.update_moving_average()
                losses.append(loss.item())
                embs = embs + emb

            st = '[{}][Epoch {}/{}] Loss: {:.4f}'.format(currentTime(),
                                                         epoch, self._args.Nepochs, np.mean(np.array(losses)))
            print(st)
            # Early Stopping Criterion
            if np.mean(np.array(losses)) < loss_t:
                loss_t = np.mean(np.array(losses))
            else:
                cnt_wait = cnt_wait + 1
            if cnt_wait > self._args.Nes:
                print("Early Stopping Criterion")
                break

        self.all_embeddings = torch.tensor(embs)
        print("\n----Node-Level Training Done! ----\n")


class NodeLevel(nn.Module):
    def __init__(self, layer_config, args):
        super().__init__()
        self._device = f'cuda:{args.device}' if torch.cuda.is_available() else "cpu"
        # encoder
        self.student_encoder = Encoder(layer_config=layer_config)
        self.teacher_encoder = copy.deepcopy(self.student_encoder)
        self.student_encoder = self.student_encoder.to(self._device)
        self.teacher_encoder = self.teacher_encoder.to(self._device)
        set_requires_grad(self.teacher_encoder, False)
        self.teacher_ema_updater = EMA(args.mad, args.Nepochs)

        rep_dim = layer_config[-1]

        # projection head
        self.student_projector = nn.Sequential(nn.Linear(rep_dim, args.N_pred_hid), nn.BatchNorm1d(args.N_pred_hid),
                                               nn.PReLU(), nn.Linear(args.N_pred_hid, rep_dim))

        self.student_projector = self.student_projector.to(self._device)

        self.student_projector.apply(init_weights)
        self.teacher_projector = copy.deepcopy(self.student_projector)
        set_requires_grad(self.teacher_projector, False)

    def reset_moving_average(self):
        del self.teacher_encoder
        self.teacher_encoder = None

    def update_moving_average(self):
        assert self.teacher_encoder is not None, 'teacher encoder has not been created yet'
        update_moving_average(self.teacher_ema_updater, self.teacher_encoder, self.student_encoder)
        update_moving_average(self.teacher_ema_updater, self.teacher_projector, self.student_projector)

    def forward(self, batch):
        student = self.student_encoder(batch.x1.to(torch.float32).to(self._device), batch.edge_index1.to(self._device))

        h1 = self.student_projector(student)

        with torch.no_grad():  # stop gradient
            teacher = self.teacher_encoder(batch.x2.to(torch.float32).to(self._device),
                                           batch.edge_index2.to(self._device))
        with torch.no_grad():  # stop gradient
            h2 = self.teacher_projector(teacher)

        emb = self.student_encoder(batch.x.to(torch.float32).to(self._device), batch.edge_index.to(self._device))

        emb = emb.detach().cpu().numpy().tolist()

        loss = get_loss(h1, h2)

        return emb, loss


class Graph_ModelTrainer(embedder):

    def __init__(self, args, block_num):
        self.block_num = block_num
        embedder.__init__(self, args)
        self._args = args
        self._init()

    def _init(self):
        args = self._args
        block_num = self.block_num
        os.environ["CUDA_VISIBLE_DEVICES"] = str(args.device)
        self._device = f'cuda:{args.device}' if torch.cuda.is_available() else "cpu"
        torch.cuda.set_device(self._device)
        if torch.cuda.is_available():
            print("using cuda")

        layers = [300] + self.hidden_layers
        # load data
        self._model = GraphLevel(layers, args).to(self._device)
        self._model.to(self._device)
        self._optimizer = optim.AdamW(params=self._model.parameters(), lr=args.Glr, weight_decay=1e-5)

        self._loader, self.true_label = get_Graph_Dataset(block_num)
        self.train()
        self.all_embeddings = F.normalize(self.all_embeddings, dim=-1, p=2).detach().cpu().numpy()

    def get_embedding(self):
        return self.all_embeddings, self.true_label

    def train(self):
        h_loss = 1e10
        cnt_wait = 0
        # Start Model Training
        print("----Graph-Level Training Start! ----\n")
        for epoch in range(self._args.Gepochs):
            losses = []
            embs = []
            for batch in self._loader:
                batch = batch.to(self._device)
                emb, loss = self._model(batch)
                self._optimizer.zero_grad()
                loss.backward()
                self._optimizer.step()
                self._model.update_moving_average()
                losses.append(loss.item())
                embs = embs + emb
            st = '[{}][Epoch {}/{}] Loss: {:.4f}'.format(currentTime(),
                                                         epoch, self._args.Gepochs, np.mean(np.array(losses)))
            print(st)

            # Early Stopping Criterion
            if np.mean(np.array(losses)) < h_loss:
                h_loss = np.mean(np.array(losses))
            elif np.mean(np.array(losses)) > h_loss:
                cnt_wait = cnt_wait + 1
            if cnt_wait > 5:
                break
        self.all_embeddings = torch.tensor(embs)
        print("\n----Graph-Level Training Done! ----")


class GraphLevel(nn.Module):
    def __init__(self, layer_config, args):
        self.args = args
        super().__init__()
        self._device = f'cuda:{args.device}' if torch.cuda.is_available() else "cpu"
        # encoder
        self.student_encoder = Encoder(layer_config)
        self.teacher_encoder = copy.deepcopy(self.student_encoder)
        self.student_encoder.to(self._device)
        self.teacher_encoder.to(self._device)
        set_requires_grad(self.teacher_encoder, False)
        self.teacher_ema_updater = EMA(args.mad, args.Gepochs)

        rep_dim = layer_config[-1]
        # projection head
        self.student_projector = nn.Sequential(nn.Linear(rep_dim, args.G_pred_hid), nn.BatchNorm1d(args.G_pred_hid),
                                               nn.PReLU(), nn.Linear(args.G_pred_hid, rep_dim))
        self.student_projector.to(self._device)
        self.student_projector.apply(init_weights)
        self.teacher_projector = copy.deepcopy(self.student_projector)
        set_requires_grad(self.teacher_projector, False)
        # pooling
        self.pool = GlobalAttention(gate_nn=nn.Sequential(
            nn.Linear(rep_dim, rep_dim), nn.BatchNorm1d(rep_dim), nn.ReLU(), nn.Linear(rep_dim, 1)))

    def reset_moving_average(self):
        del self.teacher_encoder
        self.teacher_encoder = None

    def update_moving_average(self):
        assert self.teacher_encoder is not None, 'teacher encoder has not been created yet'
        update_moving_average(self.teacher_ema_updater, self.teacher_encoder, self.student_encoder)
        update_moving_average(self.teacher_ema_updater, self.teacher_projector, self.student_projector)

    def forward(self, batch):
        student = self.student_encoder(batch.x1.to(self._device), batch.edge_index1.to(self._device))

        h1 = self.pool(student, batch.batch)
        h1 = self.student_projector(h1)

        with torch.no_grad():  # stop gradient
            teacher = self.teacher_encoder(batch.x2.to(self._device), batch.edge_index2.to(self._device))
        h2 = self.pool(teacher, batch.batch)
        with torch.no_grad():  # stop gradient
            h2 = self.teacher_projector(h2)

        emb = self.student_encoder(batch.x.to(self._device), batch.edge_index.to(self._device))
        emb = self.pool(emb, batch.batch)
        emb = emb.detach().cpu().numpy().tolist()
        loss = get_loss(h1, h2)
        res_emb = emb

        return res_emb, loss


def make_transition(trans, *items):
    transition = {}
    for key, item in zip(trans, items):
        if isinstance(item, list):
            item = torch.stack(item)
            transition[key] = item
        elif isinstance(item, np.ndarray):
            item = torch.from_numpy(item)
            transition[key] = item
        elif isinstance(item, torch.Tensor):
            transition[key] = item
        else:
            transition[key] = torch.Tensor([item])

    return transition


def make_batch(state, action, old_log_prob, advantage, old_value, learn_size, batch_size, use_cuda):
    batch = []
    total_indices = torch.randperm(learn_size)
    for i in range(learn_size // batch_size):
        indices = total_indices[batch_size * i: batch_size * (i + 1)]
        mini_state = torch.Tensor([])
        mini_action = torch.Tensor([])
        mini_old_log_prob = torch.Tensor([])
        mini_advantage = torch.Tensor([])
        mini_old_value = torch.Tensor([])
        if use_cuda:
            mini_state = mini_state.cuda()
            mini_action = mini_action.cuda()
            mini_old_log_prob = mini_old_log_prob.cuda()
            mini_advantage = mini_advantage.cuda()
            mini_old_value = mini_old_value.cuda()
        for ind in indices:
            mini_state = torch.cat((mini_state, state[ind].unsqueeze(0)), dim=0)
            mini_action = torch.cat((mini_action, action[ind].unsqueeze(0)), dim=0)
            mini_old_log_prob = torch.cat((mini_old_log_prob, old_log_prob[ind].unsqueeze(0)), dim=0)
            mini_advantage = torch.cat((mini_advantage, advantage[ind].unsqueeze(0)), dim=0)
            mini_old_value = torch.cat((mini_old_value, old_value[ind].unsqueeze(0)), dim=0)
        batch.append([mini_state, mini_action, mini_old_log_prob, mini_advantage, mini_old_value])
    return batch


def calculate_nature_cnn_out_dim(height, weight):
    size_h = np.floor((height - 8) / 4) + 1
    size_h = np.floor((size_h - 4) / 2) + 1
    size_h = np.floor((size_h - 3) / 1) + 1
    size_w = np.floor((weight - 8) / 4) + 1
    size_w = np.floor((size_w - 4) / 2) + 1
    size_w = np.floor((size_w - 3) / 1) + 1
    return size_h, size_w


class DQN_Config:
    def __init__(self, input_type, input_size=None):
        self.max_buffer = 100000
        self.update_freq = 200
        self.use_cuda = True
        self.trans = ["state", "action", "reward", "next_state", "done"]
        self.lr = 0.001
        self.tau = 0.005
        self.gamma = 0.99
        self.batch_size = 128
        self.max_grad_norm = 1
        self.epsilon_init = 1
        self.epsilon_decay = 0.995
        self.epsilon_min = 0.05
        self.q_layer = [256, 256]
        if input_type == "vector":
            self.encoder = "mlp"
            self.encoder_layer = [512, 256]
            self.feature_dim = 256
        elif input_type == "image":
            self.encoder = "cnn"
            self.encoder_layer = [[input_size[0], 32, 8, 4],
                                  [32, 64, 4, 2],
                                  [64, 64, 3, 1]]
            size_h, size_w = calculate_nature_cnn_out_dim(input_size[1], input_size[2])
            self.feature_dim = [int(64 * size_h * size_w), 256]


class DQN:
    def __init__(self, state_dim, action_dim, input_type="vector", args=None):
        if args is None:
            self.args = DQN_Config(input_type, state_dim)
        self.action_dim = action_dim
        self.buffer = BaseBuffer(self.args.trans, self.args.max_buffer)
        self.policy_net = QNet(state_dim[0], action_dim, self.args.q_layer, self.args.encoder, self.args.encoder_layer,
                               self.args.feature_dim)
        self.target_net = QNet(state_dim[0], action_dim, self.args.q_layer, self.args.encoder, self.args.encoder_layer,
                               self.args.feature_dim)
        if self.args.use_cuda:
            self.policy_net = self.policy_net.cuda()
            self.target_net = self.target_net.cuda()
        self.update_network()
        self.policy_net.eval()
        self.target_net.eval()
        self.optimizer = optim.Adam(self.policy_net.parameters(), lr=self.args.lr)

        self.epsilon = self.args.epsilon_init

    def select_action(self, state, epsilon=None):
        if epsilon is None:
            epsilon = self.epsilon
        if random.random() > epsilon:
            state = torch.Tensor(state)
            if self.args.use_cuda:
                state = state.cuda()
            q_value = self.policy_net(state)
            return torch.argmax(q_value).cpu().unsqueeze(0).detach()
        else:
            return torch.Tensor([random.choice(np.arange(self.action_dim))]).type(torch.int64).detach()

    def add_buffer(self, transition):
        self.buffer.add(transition)

    def epsilon_decay(self):
        self.epsilon = max(self.epsilon * self.args.epsilon_decay, self.args.epsilon_min)

    def update_network(self):
        self.target_net.load_state_dict(self.policy_net.state_dict())

    def save_model(self, model_path):
        torch.save(self.policy_net.state_dict(), model_path)

    def learn(self, step):
        self.policy_net.train()
        data = self.buffer.sample(self.args.batch_size)
        states = torch.stack(data["state"])  # [batch_size, state_dim]
        actions = torch.stack(data["action"])  # [batch_size, 1]
        rewards = torch.stack(data["reward"])  # [batch_size, 1]
        next_states = torch.stack(data["next_state"])  # [batch_size, state_dim]
        dones = torch.stack(data["done"])  # [batch_size, 1]
        # print("shape check", states.shape, actions.shape, rewards.shape, next_states.shape, dones.shape)
        if self.args.use_cuda:
            states = states.cuda()
            actions = actions.cuda()
            rewards = rewards.cuda()
            next_states = next_states.cuda()
            dones = dones.cuda()
        actions = actions.type(torch.int64)
        rewards = (rewards - rewards.mean()) / (rewards.std() + 1e-7)

        # get q-values for all actions in current states
        predicted_q = self.policy_net(states)  # [batch_size, action_dim]
        # select q-values for chosen actions
        predicted_q_actions = torch.gather(predicted_q, -1, actions)  # [batch_size, 1]
        # compute q-values for all actions in next states
        predicted_next_q = self.target_net(next_states)  # [batch_size, action_dim]
        # compute V*(next_states) using predicted next q-values
        next_state_values, indexes = torch.max(predicted_next_q, dim=-1)  # [batch_size]
        next_state_values = next_state_values.unsqueeze(-1)  # [batch_size, 1]
        # compute "target q-values" for loss - it's what's inside square parentheses in the above formula.
        target_q_actions = rewards + self.args.gamma * next_state_values * (1 - dones)  # [batch_size, 1]
        loss = nn.SmoothL1Loss()(predicted_q_actions, target_q_actions.detach())

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        self.policy_net.eval()
        self.epsilon_decay()

        if step % self.args.update_freq == 0:
            print("update target network")
            self.update_network()
        return loss.item()


class PPO_Config:
    def __init__(self, input_type, input_size=None):
        self.max_buffer = 2048
        self.trainable_std = False
        self.use_cuda = True
        self.trans = ["state", "action", "reward", "done", "log_prob"]
        self.lr = 0.0003
        self.gamma = 0.99
        self.lambda_ = 0.95

        self.train_epoch = 80
        self.clip_ratio = 0.2
        self.critic_coef = 0.5
        self.entropy_coef = 0.01
        self.max_grad_norm = 0.5

        self.action_std_init = 0.6
        self.action_std_decay_rate = 0.05
        self.action_std_min = 0.1
        self.action_std_update_freq = 100

        self.actor_layer = [32, 32]
        self.critic_layer = [32, 32]
        if input_type == "vector":
            self.encoder = "mlp"
            self.encoder_layer = [64, 64]
            self.feature_dim = 32
        elif input_type == "image":
            self.encoder = "cnn"
            self.encoder_layer = [[input_size[0], 32, 8, 4],
                                  [32, 64, 4, 2],
                                  [64, 64, 3, 1]]
            size_h, size_w = calculate_nature_cnn_out_dim(input_size[1], input_size[2])
            self.feature_dim = [int(64 * size_h * size_w), 256]


class PPO:
    def __init__(self, state_dim, action_dim, continuous=True, input_type="vector", args=None):
        if args is None:
            self.args = PPO_Config(input_type, state_dim)
        self.buffer = BaseBuffer(self.args.trans, self.args.max_buffer)
        self.model = ActorCritic(state_dim[0], action_dim, self.args.actor_layer, self.args.critic_layer,
                                 self.args.encoder, self.args.encoder_layer, self.args.feature_dim, continuous,
                                 self.args.action_std_init)
        if self.args.use_cuda:
            self.model = self.model.cuda()
        self.model.eval()
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.args.lr)

    def select_action(self, state):
        state = torch.Tensor(state).float()
        if self.args.use_cuda:
            state = state.cuda()
        action, action_log_prob = self.model.act(state)
        return action.detach().cpu(), action_log_prob.detach().cpu()

    def add_buffer(self, transition):
        self.buffer.add(transition)

    def save_model(self, model_path):
        torch.save(self.model.state_dict(), model_path)

    def learn(self):
        self.model.train()
        data, size = self.buffer.get_data_buffer()
        states = data["state"]
        actions = data["action"]
        rewards = data["reward"]
        dones = data["done"]
        old_log_probs = data["log_prob"]

        # Monte Carlo estimate of returns
        returns = []
        discounted_reward = 0
        for reward, done in zip(reversed(rewards), reversed(dones)):
            if done:
                discounted_reward = 0
            discounted_reward = reward + (self.args.gamma * discounted_reward)
            returns.insert(0, discounted_reward)
        # Normalizing the rewards
        returns = torch.tensor(returns, dtype=torch.float32)
        returns = (returns - returns.mean()) / (returns.std() + 1e-5)

        # trans into tensor and send to cpu/gpu
        states = torch.stack(states).float()
        actions = torch.stack(actions)
        old_log_probs = torch.stack(old_log_probs)
        if self.args.use_cuda:
            states = states.cuda()  # [batch_size, state_dim]
            actions = actions.cuda()  # [batch_size]
            old_log_probs = old_log_probs.cuda()  # [batch_size]
            returns = returns.cuda()  # [batch_size]

        loss_list = []
        for e in range(self.args.train_epoch):
            # Evaluating old actions and values
            log_probs, values, dist_entropy = self.model.evaluate_AC(states, actions)
            # print("shape check", log_probs.shape, values.shape, dist_entropy.shape)
            values = values.squeeze(-1)

            # Finding the ratio (pi_theta / pi_theta__old)
            ratios = torch.exp(log_probs - old_log_probs.detach())

            # Finding Surrogate Loss
            advantages = returns - values.detach()
            surr1 = ratios * advantages
            surr2 = torch.clamp(ratios, 1 - self.args.clip_ratio, 1 + self.args.clip_ratio) * advantages

            # final loss of clipped objective PPO
            actor_loss = -torch.min(surr1, surr2)
            critic_loss = nn.MSELoss()(values, returns)
            entropy_bonus = -dist_entropy
            loss = actor_loss + 0.5 * critic_loss + 0.01 * entropy_bonus
            loss = loss.mean()

            # take gradient step
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            loss_list.append(loss.item())
        self.model.eval()
        self.buffer.clear()
        return np.mean(loss_list)


class ActorCritic(nn.Module):
    def __init__(self, state_dim, action_dim, actor_layer, critic_layer, encoder=None, encoder_layer=None,
                 feature_dim=None, continuous=False, std_init=0.6):
        super(ActorCritic, self).__init__()
        self.continuous = continuous
        if continuous:
            self.action_var = torch.full((action_dim,), std_init * std_init)

        self.encoder = encoder
        if encoder is None:
            input_dim = state_dim
        elif encoder == "mlp":
            self.encoder = MLPEncoder(state_dim, encoder_layer, feature_dim)
            input_dim = self.encoder.get_dim()
        elif encoder == "cnn":
            self.encoder = CNNEncoder(encoder_layer, feature_dim)
            input_dim = self.encoder.get_dim()
        else:
            raise NotImplementedError

        if self.continuous:
            layers = [nn.Linear(input_dim, actor_layer[0]),
                      nn.ReLU(inplace=True)]
            for i in range(len(actor_layer) - 1):
                layers.append(nn.Linear(actor_layer[i], actor_layer[i + 1]))
                layers.append(nn.ReLU(inplace=True))
            layers.append(nn.Linear(actor_layer[-1], action_dim))
            self.actor = nn.Sequential(*layers)

        else:
            layers = [nn.Linear(input_dim, actor_layer[0]),
                      nn.ReLU(inplace=True)]
            for i in range(len(actor_layer) - 1):
                layers.append(nn.Linear(actor_layer[i], actor_layer[i + 1]))
                layers.append(nn.ReLU(inplace=True))
            layers.append(nn.Linear(actor_layer[-1], action_dim))
            layers.append(nn.Softmax(dim=-1))
            self.actor = nn.Sequential(*layers)

        layers = [nn.Linear(input_dim, critic_layer[0]),
                  nn.ReLU(inplace=True)]
        for i in range(len(critic_layer) - 1):
            layers.append(nn.Linear(critic_layer[i], critic_layer[i + 1]))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(critic_layer[-1], 1))
        self.critic = nn.Sequential(*layers)
        self.__network_init()

    def forward(self):
        raise NotImplementedError

    def __network_init(self):
        for layer in self.modules():
            if isinstance(layer, nn.Linear):
                nn.init.orthogonal_(layer.weight)
                if layer.bias is not None:
                    layer.bias.data.zero_()

    def act(self, state):
        if self.encoder is not None:
            state = self.encoder(state)
        if self.continuous:
            mu = self.actor(state).cpu()
            cov_mat = torch.diag(self.action_var)
            dist = MultivariateNormal(mu, cov_mat)
        else:
            action_prob = self.actor(state)
            dist = Categorical(action_prob)
        action = dist.sample()
        action_log_prob = dist.log_prob(action)
        return action.detach(), action_log_prob.detach()

    def evaluate_AC(self, state, action):
        if self.encoder is not None:
            state = self.encoder(state)
        if self.continuous:
            mu = self.actor(state).cpu()
            action_var = self.action_var.expand_as(mu)
            cov_mat = torch.diag_embed(action_var)
            dist = MultivariateNormal(mu, cov_mat)
        else:
            action_prob = self.actor(state)
            dist = Categorical(action_prob)
        action_log_prob = dist.log_prob(action.cpu())
        dist_entropy = dist.entropy()
        value = self.critic(state)
        return action_log_prob.cuda(), value, dist_entropy.cuda()


class MLPEncoder(nn.Module):
    def __init__(self, state_dim, layer_dim: list, feature_dim: int):
        super(MLPEncoder, self).__init__()
        layers = [nn.Linear(state_dim, layer_dim[0]),
                  # nn.BatchNorm1d(layer_dim[0]),
                  nn.ReLU(inplace=True)]
        for i in range(len(layer_dim) - 1):
            layers.append(nn.Linear(layer_dim[i], layer_dim[i + 1]))
            # layers.append(nn.BatchNorm1d(layer_dim[i + 1]))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(layer_dim[-1], feature_dim, bias=False))
        self.encoder = nn.Sequential(*layers)
        self.out_dim = feature_dim

    def forward(self, x):
        return self.encoder(x)

    def get_dim(self):
        return self.out_dim


class CNNEncoder(nn.Module):
    def __init__(self, layer_dim: list, feature_dim: list):
        super(CNNEncoder, self).__init__()
        layers = []
        for layer in layer_dim:
            layers.append(nn.Conv2d(layer[0], layer[1], layer[2], layer[3]))
            # layers.append(nn.BatchNorm2d(layer[1]))
            layers.append(nn.ReLU(inplace=True))
        self.encoder = nn.Sequential(*layers)
        self.projector = nn.Linear(feature_dim[0], feature_dim[1], bias=False)
        self.out_dim = feature_dim[1]

    def forward(self, x):
        f = self.encoder(x)
        f = f.reshape(f.shape[0], -1)
        f = self.projector(f)
        return f

    def get_dim(self):
        return self.out_dim


class QNet(nn.Module):
    def __init__(self, state_dim, action_dim, q_layer, encoder=None, encoder_layer=None, feature_dim=None):
        super(QNet, self).__init__()
        self.encoder = encoder
        if encoder is None:
            input_dim = state_dim
        elif encoder == "mlp":
            self.encoder = MLPEncoder(state_dim, encoder_layer, feature_dim)
            input_dim = self.encoder.get_dim()
        elif encoder == "cnn":
            self.encoder = CNNEncoder(encoder_layer, feature_dim)
            input_dim = self.encoder.get_dim()
        else:
            raise NotImplementedError

        layers = [nn.Linear(input_dim, q_layer[0]),
                  nn.ReLU(inplace=True)]
        for i in range(len(q_layer) - 1):
            layers.append(nn.Linear(q_layer[i], q_layer[i + 1]))
            layers.append(nn.ReLU(inplace=True))
        layers.append(nn.Linear(q_layer[-1], action_dim))
        self.q_net = nn.Sequential(*layers)
        self.__network_init()

    def __network_init(self):
        for layer in self.modules():
            if isinstance(layer, nn.Linear):
                nn.init.orthogonal_(layer.weight)
                if layer.bias is not None:
                    layer.bias.data.zero_()

    def forward(self, state):
        if self.encoder is not None:
            state = self.encoder(state)
        q = self.q_net(state)
        return q


class BaseBuffer:
    def __init__(self, trans, max_len):
        self.trans = trans
        self.max_len = max_len
        self.data = {}
        for key in trans:
            self.data[key] = deque(maxlen=self.max_len)
        self.total_idx = 0

    def get_len(self):
        return self.total_idx

    def clear(self):
        """
        clear the buffer
        :return:
        """
        self.data = {}
        for key in self.trans:
            self.data[key] = []
        self.total_idx = 0

    def add(self, transition):
        """
        add a transition in buffer
        :return:
        """
        for key in transition:
            self.data[key].append(transition[key])
        self.total_idx += 1

    def get_data_buffer(self):
        data_size = len(self.data["state"])
        data = {}
        for key in self.data:
            data[key] = list(self.data[key])
        return data, data_size

    def sample(self, size):
        data_size = len(self.data["state"])
        size = min(data_size, size)
        indices = torch.randperm(data_size)[:size]
        data = {}
        for key in self.trans:
            data[key] = []
            for idx, ind in enumerate(indices):
                data[key].append(self.data[key][ind])
        return data


def unique(lists):
    # delete duplicate attribute values
    lists = list(map(lambda x: x.lower(), lists))
    if lists[0] == '':
        res = []
    else:
        res = [lists[0]]
    for i in range(len(lists)):
        if i == 0 or (lists[i] in lists[0:i]) or lists[i] == '':
            continue
        else:
            res.append(lists[i])
    return res


def construct_graph_from_df(df, G=None):
    # construct graph according to df
    if G is None:
        G = nx.Graph()
    for _, row in df.iterrows():
        tid = 't_' + str(row['tweet_id'])
        G.add_node(tid)

        user_ids = row['user_mentions']
        user_ids.append(row['user_id'])
        user_ids = ['u_' + str(each) for each in user_ids]
        G.add_nodes_from(user_ids)

        words = row['filtered_words']
        words = [('w_' + each).lower() for each in words]
        G.add_nodes_from(words)

        hashtags = row['hashtags']
        hashtags = [('h_' + each).lower() for each in hashtags]
        G.add_nodes_from(hashtags)

        edges = []
        # Connect the message node with each related user node, word node, etc
        edges += [(tid, each) for each in user_ids]
        edges += [(tid, each) for each in words]
        edges += [(tid, each) for each in hashtags]
        G.add_edges_from(edges)
    return G


def construct_graph(data, feature, index):
    # Build graph for a single tweet
    G = nx.Graph()
    X = []

    tweet = data["text"].values
    X.append(feature[index].tolist())
    index = index + 1
    tweet_id = data["tweet_id"].values
    G.add_node(tweet_id[0])

    f_w = data["filtered_words"].tolist()
    edges = []

    h_t = data["hashtags"].tolist()
    h_t = h_t[0]
    n =  + f_w[0] + h_t
    n = unique(n)

    if len(n) != 0:
        for each in n:
            X.append(feature[index].tolist())
            index = index + 1
        G.add_nodes_from(n)
        edges += [(tweet_id[0], each) for each in n]
    G.add_edges_from(edges)
    return G, X


def normalize_adj(adj):
    # Symmetrically normalize adjacency matrix
    adj = sp.coo_matrix(adj)
    rowsum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    return adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()


def aug_edge(adj):  # edge perturbation
    adj = np.array(adj)
    aug_adj1 = np.array([[i for i in j] for j in adj])
    aug_adj2 = np.array([[i for i in j] for j in adj])
    p = np.random.randint(0, len(adj) - 1)
    aug_adj1[p][0] = 0
    aug_adj1[0][p] = 0
    t = np.random.randint(1, len(adj) - 1)
    aug_adj1[t][p] = 1
    aug_adj1[p][t] = 1

    p = np.random.randint(0, len(adj) - 1)
    aug_adj2[p][0] = 0
    aug_adj2[0][p] = 0
    t = np.random.randint(1, len(adj) - 1)
    aug_adj2[t][p] = 1
    aug_adj2[p][t] = 1

    return aug_adj1, aug_adj2


def get_edge_index(adj):  # Get edge set according to adjacency matrix

    edge_index1 = []
    edge_index2 = []
    for i in range(len(adj)):
        for j in range(len(adj)):
            if adj[i][j] == 1 and i < j:
                edge_index1.append(i)
                edge_index2.append(j)

    edge_index = [edge_index1] + [edge_index2]

    return edge_index


def get_data(message_num, start, tweet_sum, save_path):
    os.makedirs(save_path, exist_ok=True)
    df = DatasetLoader("arabic_twitter").load_data()
    df = df.sort_values(by='created_at').reset_index()
    ini_df = df[start:tweet_sum]

    G = construct_graph_from_df(ini_df)

    d_features = documents_to_features(df)
    print("Document features generated.")
    t_features = df_to_t_features(df)
    print("Time features generated.")
    combined_features = np.concatenate((d_features, t_features), axis=1)
    print("Concatenated document features and time features.")
    np.save(save_path + 'features_69612_0709_spacy_lg_zero_multiclasses_filtered.npy', combined_features)
    print("Initial features saved.")

    combined_features = np.load(save_path + 'features_69612_0709_spacy_lg_zero_multiclasses_filtered.npy')
    A = nx.adjacency_matrix(G).todense().tolist()

    X = []
    nodes = list(G.nodes)

    tweet = []
    j = 0

    for i in range(len(nodes)):
        t = nodes[i][0:2]
        e = nodes[i][2:]
        if t == "t_":
            tweet.append(i)
            index = list(ini_df["tweet_id"]).index(int(e))
            X.append(list(combined_features[index]))
            j = j + 1
    X = torch.tensor(X)
    adj = np.array([[0] * len(tweet)] * len(tweet))

    for i in range(len(tweet)):
        for j in range(len(A)):
            if A[tweet[i]][j] == 1:
                for s in range(len(tweet)):
                    if A[j][tweet[s]] == 1 and s != i:
                        adj[i][s] = 1
    edge_index = get_edge_index(adj)

    edge_index1 = copy.deepcopy(edge_index)
    edge_index2 = copy.deepcopy(edge_index)
    true_y = torch.tensor(list(ini_df['event_id']))

    drop_percent = 0.2
    i = 0
    while 1:
        if i >= len(G.edges) * drop_percent:
            break
        m1 = random.randint(0, len(edge_index1[0]) - 1)
        m2 = random.randint(0, len(edge_index2[0]) - 1)
        if m1 == m2:
            continue
        else:
            del edge_index1[0][m1]
            del edge_index1[1][m1]
            del edge_index2[0][m2]
            del edge_index2[1][m2]

        i = i + 1
    edge_index = torch.tensor(edge_index)
    edge_index1 = torch.tensor(edge_index1)
    edge_index2 = torch.tensor(edge_index2)

    dict_graph = {}

    dict_graph['x'] = X
    dict_graph['x1'] = X
    dict_graph['x2'] = X
    dict_graph['edge_index'] = edge_index
    dict_graph['edge_index1'] = edge_index1
    dict_graph['edge_index2'] = edge_index2
    dict_graph['y'] = true_y
    return dict_graph


def getData(M_num):  # construct an entire graph within a block
    #     print(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time())))
    args = args_define.args
    M = [20254, 28976, 30467, 32302, 34312, 36146, 37422, 42700, 44260, 45623, 46719,
         47951, 51188, 53160, 56116, 58665, 59575, 62251, 64138, 65537, 66430, 68840]

    if M_num == 0:
        num = 0
        size = 500
    elif M_num == 1:
        num = M[M_num - 1]
        size = 500
    elif M[M_num] - M[M_num - 1] > 2000:
        num = M[M_num - 1]
        size = 1000
    else:
        num = M[M_num - 1]
        size = M[M_num] - M[M_num - 1]
    data = []
    i = M_num
    j = 0

    while 1:
        if (num + size) >= M[i]:
            tmp = get_data(i, num, M[i], args.file_path)
            data.append(tmp)
            break
        else:
            tmp = get_data(i, num, num + size, args.file_path)
            data.append(tmp)
            j = j + 1
            print("***************Block " + str(j) + " is done.****************")
            num = num + size

    save_data(data, args.file_path, M_num)
    return data


def save_data(data, save_path, M_num):
    os.makedirs(save_path, exist_ok=True)
    file_path = os.path.join(save_path, f'data_{M_num}.pkl')
    with open(file_path, 'wb') as f:
        pickle.dump(data, f)
    print(f"Data saved at {file_path}")


def get_Graph_Dataset(message_number):
    args = args_define().args
    print("\nBuilding graph-level social network...")
    start_time = time.time()
    # load data for graph-level contrastive learning
    dataset = []
    label = []
    file_name = args.file_path + 'GCL-data/message_block_' + str(message_number) + '.npy'
    data = np.load(file_name, allow_pickle=True)

    for dict_data in data:
        data = Data(x=dict_data['X'], x1=dict_data['x1'], x2=dict_data['x2'],
                    edge_index=dict_data['edge_index'], edge_index1=dict_data['edge_index1'],
                    edge_index2=dict_data['edge_index2'])
        dataset.append(data)
        label.append(dict_data['label'])
    if message_number == 0:
        dataset = loader.DataLoader(dataset, batch_size=4096)
    else:
        dataset = loader.DataLoader(dataset, batch_size=len(dataset))
    end_time = time.time()
    run_time = end_time - start_time
    print("Done! It takes " + str(int(run_time)) + " seconds.\n")
    return dataset, label


def get_Node_Dataset(message_number):
    # load data for node-level contrastive learning
    print("\nBuilding node-level social network...")
    start_time = time.time()
    # datas = getData(message_number)
    file_path = os.path.join(args.file_path, f'data_{message_number}.pkl')

    if os.path.exists(file_path):
        with open(file_path, 'rb') as f:
            datas = pickle.load(f)
        print("Data loaded successfully.")
        # 现在你可以使用 data 进行进一步操作
    else:
        print(f"No data file found at {file_path}")
        datas = getData(message_number)

    dataset = []
    labels = []

    for data in datas:
        dict_data = data

        dict_data['x'] = torch.tensor(np.array(dict_data['x']))
        dict_data['x1'] = torch.tensor(np.array(dict_data['x1']))
        dict_data['x2'] = torch.tensor(np.array(dict_data['x2']))
        dict_data['edge_index'] = torch.tensor(np.array(dict_data['edge_index']))
        dict_data['edge_index1'] = torch.tensor(np.array(dict_data['edge_index1']))
        dict_data['edge_index2'] = torch.tensor(np.array(dict_data['edge_index2']))
        data = Data(x=dict_data['x'], x1=dict_data['x1'], x2=dict_data['x2'],
                    edge_index=dict_data['edge_index'], edge_index1=dict_data['edge_index1'],
                    edge_index2=dict_data['edge_index2'])

        label = dict_data['y']
        if len(labels) == 0:
            labels = label
        else:
            labels = torch.cat([labels, label])

        dataset.append(data)
    end_time = time.time()
    run_time = end_time - start_time
    print("Done! It takes " + str(int(run_time)) + " seconds.\n")
    return dataset, np.array(labels).tolist()


# Calculate the embeddings of all the documents in the dataframe,
# the embedding of each document is an average of the pre-trained embeddings of all the words in it
def documents_to_features(df):
    nlp = en_core_web_lg.load()
    features = df.filtered_words.apply(lambda x: nlp(' '.join(x)).vector).values
    return np.stack(features, axis=0)


# encode one times-tamp
# t_str: a string of format '2012-10-11 07:19:34'
def extract_time_feature(t_str):
    t = datetime.fromisoformat(str(t_str))
    OLE_TIME_ZERO = datetime(1899, 12, 30)
    delta = t - OLE_TIME_ZERO
    return [(float(delta.days) / 100000.), (float(delta.seconds) / 86400)]  # 86,400 seconds in day


# encode the times-tamps of all the messages in the dataframe
def df_to_t_features(df):
    t_features = np.asarray([extract_time_feature(t_str) for t_str in df['created_at']])
    return t_features


if __name__ == "__main__":
    args = args_define().args
    dataset = DatasetLoader("arabic_twitter").load_data()
    hcrc = HCRC(args, dataset)

    predictions, ground_truths = hcrc.detection()  # 进行预测
    results = hcrc.evaluate(predictions, ground_truths)  # 评估模型



import pytest
import torch
import numpy as np
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from detector.UCLSED import *
# 创建一个虚拟的DatasetLoader实例
class MockDatasetLoader:
    def load_data(self):
        return pd.DataFrame({
            'tweet_id': [1, 2, 3],
            'user_mentions': [['user1'], ['user2'], ['user3']],
            'text': ['text1', 'text2', 'text3'],
            'hashtags': [['hashtag1'], ['hashtag2'], ['hashtag3']],
            'entities': [['entity1'], ['entity2'], ['entity3']],
            'urls': [['url1'], ['url2'], ['url3']],
            'filtered_words': [['word1'], ['word2'], ['word3']],
            'created_at': ['2023-01-01', '2023-01-02', '2023-01-03'],
            'event_id': [0, 1, 0]
        })

# 替换DatasetLoader为MockDatasetLoader
@pytest.fixture
def mock_dataset():
    return MockDatasetLoader().load_data()

# 测试UCLSED类的初始化
def test_uclsed_init(mock_dataset):
    args = args_define().args
    uclsed = UCLSED(args, mock_dataset)
    assert uclsed.save_path is None
    assert uclsed.test_indices is None
    assert uclsed.val_indices is None
    assert uclsed.train_indices is None
    assert uclsed.mask_path is None
    assert uclsed.labels is None
    assert uclsed.times is None
    assert uclsed.g_dict is None
    assert uclsed.views is None
    assert uclsed.features is None

# 测试Preprocessor类的初始化
def test_preprocessor_init(mock_dataset):
    preprocessor = Preprocessor(mock_dataset)
    assert preprocessor is not None

# 测试Preprocessor类的load_data方法
def test_preprocessor_load_data(mock_dataset):
    preprocessor = Preprocessor(mock_dataset)
    event_df = preprocessor.load_data(mock_dataset)
    assert isinstance(event_df, pd.DataFrame)
    assert event_df.shape == (3, 9)

# 测试Preprocessor类的get_nlp方法
def test_preprocessor_get_nlp():
    preprocessor = Preprocessor(None)
    nlp = preprocessor.get_nlp("English")
    assert nlp is not None
    nlp = preprocessor.get_nlp("French")
    assert nlp is not None

# 测试Preprocessor类的construct_graph方法
def test_preprocessor_construct_graph(mock_dataset, tmpdir):
    args = args_define().args
    args.file_path = str(tmpdir) + '/'
    preprocessor = Preprocessor(mock_dataset)
    preprocessor.construct_graph(mock_dataset)
    assert os.path.exists(str(tmpdir) + '/features.npy')
    assert os.path.exists(str(tmpdir) + '/time.npy')
    assert os.path.exists(str(tmpdir) + '/label.npy')

# 测试UCLSED类的fit方法
def test_uclsed_fit(mock_dataset, tmpdir):
    args = args_define().args
    args.file_path = str(tmpdir) + '/'
    args.save_path = str(tmpdir) + '/'
    uclsed = UCLSED(args, mock_dataset)
    uclsed.fit()
    assert os.path.exists(str(tmpdir) + '/train_indices.pt')
    assert os.path.exists(str(tmpdir) + '/val_indices.pt')
    assert os.path.exists(str(tmpdir) + '/test_indices.pt')

# 测试UCLSED类的detection方法
def test_uclsed_detection(mock_dataset, tmpdir):
    args = args_define().args
    args.file_path = str(tmpdir) + '/'
    args.save_path = str(tmpdir) + '/'
    uclsed = UCLSED(args, mock_dataset)
    uclsed.fit()
    ground_truth, predictions = uclsed.detection()
    assert isinstance(ground_truth, np.ndarray)
    assert isinstance(predictions, np.ndarray)

# 测试UCLSED类的evaluate方法
def test_uclsed_evaluate(mock_dataset, tmpdir):
    args = args_define().args
    args.file_path = str(tmpdir) + '/'
    args.save_path = str(tmpdir) + '/'
    uclsed = UCLSED(args, mock_dataset)
    uclsed.fit()
    ground_truth, predictions = uclsed.detection()
    val_f1, val_acc = uclsed.evaluate(ground_truth, predictions)
    assert isinstance(val_f1, float)
    assert isinstance(val_acc, float)


import pytest
import torch
import numpy as np
from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, normalized_mutual_info_score
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from detector.KPGNN import KPGNN, args_define, SocialDataset

tmpdir="../model/model_saved/kpgnn/kpgnn_incremental_test/"

# 模拟数据集
class MockDataset:
    def __init__(self):
        self.features = np.random.rand(100, 300)
        self.labels = np.random.randint(0, 10, 100)
        self.matrix = np.random.rand(100, 100)

# 测试 KPGNN 类的初始化
def test_kpgnn_init():
    args = args_define().args
    dataset = MockDataset()
    kpgnn = KPGNN(args, dataset)
    assert kpgnn.args == args
    assert kpgnn.dataset == dataset
    assert kpgnn.model is None
    assert kpgnn.loss_fn is None
    assert kpgnn.loss_fn_dgi is None
    assert kpgnn.metrics is None
    assert kpgnn.train_indices is None
    assert kpgnn.indices_to_remove is None
    assert kpgnn.embedding_save_path is None
    assert kpgnn.data_split is None

# 测试 KPGNN 类的 preprocess 方法
def test_kpgnn_preprocess(tmpdir):
    args = args_define().args
    args.data_path = str(tmpdir)
    dataset = MockDataset()
    kpgnn = KPGNN(args, dataset)
    kpgnn.preprocess()
    # 这里可以添加更多的断言来检查预处理后的数据

# 测试 KPGNN 类的 fit 方法
def test_kpgnn_fit(tmpdir):
    args = args_define().args
    args.data_path = str(tmpdir)
    dataset = MockDataset()
    kpgnn = KPGNN(args, dataset)
    kpgnn.fit()
    # 这里可以添加更多的断言来检查模型是否正确训练

# 测试 KPGNN 类的 detection 方法
def test_kpgnn_detection(tmpdir):
    args = args_define().args
    args.data_path = str(tmpdir)
    dataset = MockDataset()
    kpgnn = KPGNN(args, dataset)
    kpgnn.fit()
    predictions, ground_truths = kpgnn.detection()
    assert isinstance(predictions, np.ndarray)
    assert isinstance(ground_truths, np.ndarray)
    assert len(predictions) == len(ground_truths)

# 测试 KPGNN 类的 evaluate 方法
def test_kpgnn_evaluate(tmpdir):
    args = args_define().args
    args.data_path = str(tmpdir)
    dataset = MockDataset()
    kpgnn = KPGNN(args, dataset)
    kpgnn.fit()
    predictions, ground_truths = kpgnn.detection()
    ars, ami, nmi = kpgnn.evaluate(predictions, ground_truths)
    assert isinstance(ars, float)
    assert isinstance(ami, float)
    assert isinstance(nmi, float)
    assert 0 <= ars <= 1
    assert 0 <= ami <= 1
    assert 0 <= nmi <= 1

# 测试 SocialDataset 类的初始化
def test_social_dataset_init(tmpdir):
    data_path = str(tmpdir.mkdir("0"))
    np.save(os.path.join(data_path, "features.npy"), np.random.rand(100, 300))
    np.save(os.path.join(data_path, "labels.npy"), np.random.randint(0, 10, 100))
    sparse.save_npz(os.path.join(data_path, "s_bool_A_tid_tid.npz"), sparse.csr_matrix(np.random.rand(100, 100)))
    dataset = SocialDataset(str(tmpdir), 0)
    assert isinstance(dataset.features, np.ndarray)
    assert isinstance(dataset.labels, np.ndarray)
    assert isinstance(dataset.matrix, sparse.csr_matrix)

# 测试 SocialDataset 类的 __len__ 方法
def test_social_dataset_len(tmpdir):
    data_path = str(tmpdir.mkdir("0"))
    np.save(os.path.join(data_path, "features.npy"), np.random.rand(100, 300))
    np.save(os.path.join(data_path, "labels.npy"), np.random.randint(0, 10, 100))
    sparse.save_npz(os.path.join(data_path, "s_bool_A_tid_tid.npz"), sparse.csr_matrix(np.random.rand(100, 100)))
    dataset = SocialDataset(str(tmpdir), 0)
    assert len(dataset) == dataset.features.shape[0]

# 测试 SocialDataset 类的 __getitem__ 方法
def test_social_dataset_getitem(tmpdir):
    data_path = str(tmpdir.mkdir("0"))
    np.save(os.path.join(data_path, "features.npy"), np.random.rand(100, 300))
    np.save(os.path.join(data_path, "labels.npy"), np.random.randint(0, 10, 100))
    sparse.save_npz(os.path.join(data_path, "s_bool_A_tid_tid.npz"), sparse.csr_matrix(np.random.rand(100, 100)))
    dataset = SocialDataset(str(tmpdir), 0)
    feature, label = dataset[0]
    assert isinstance(feature, np.ndarray)
    assert isinstance(label, np.ndarray)

# 测试 SocialDataset 类的 load_adj_matrix 方法
def test_social_dataset_load_adj_matrix(tmpdir):
    data_path = str(tmpdir.mkdir("0"))
    np.save(os.path.join(data_path, "features.npy"), np.random.rand(100, 300))
    np.save(os.path.join(data_path, "labels.npy"), np.random.randint(0, 10, 100))
    sparse.save_npz(os.path.join(data_path, "s_bool_A_tid_tid.npz"), sparse.csr_matrix(np.random.rand(100, 100)))
    dataset = SocialDataset(str(tmpdir), 0)
    assert isinstance(dataset.matrix, sparse.csr_matrix)

# 测试 SocialDataset 类的 remove_obsolete_nodes 方法
def test_social_dataset_remove_obsolete_nodes(tmpdir):
    data_path = str(tmpdir.mkdir("0"))
    np.save(os.path.join(data_path, "features.npy"), np.random.rand(100, 300))
    np.save(os.path.join(data_path, "labels.npy"), np.random.randint(0, 10, 100))
    sparse.save_npz(os.path.join(data_path, "s_bool_A_tid_tid.npz"), sparse.csr_matrix(np.random.rand(100, 100)))
    dataset = SocialDataset(str(tmpdir), 0)
    original_length = len(dataset)
    dataset.remove_obsolete_nodes([0, 1, 2])
    assert len(dataset) == original_length - 3

# 运行测试
if __name__ == "__main__":
    pytest.main()


import os
import numpy as np
import pandas as pd
from os.path import exists
import pickle
import torch
from sentence_transformers import SentenceTransformer
import re
from sklearn.metrics.cluster import normalized_mutual_info_score, adjusted_mutual_info_score, adjusted_rand_score
from sklearn import metrics
from itertools import combinations, chain
import networkx as nx
from datetime import datetime
import math
from networkx.algorithms import cuts
from sklearn.model_selection import train_test_split
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader, DatasetLoader_mini




#2012 user_id,user_mentions,hashtags,entities,text,event_id
class HISEvent():
    def __init__(self, dataset, language):
        self.dataset = dataset
        self.language = language
        self.save_path = "../model/model_saved/hisevent/Event2012_mini/"

    def preprocess(self):
        preprocessor = Preprocessor(self.dataset)
        preprocessor.preprocess_data(self.save_path, self.dataset, self.language)

    def detection(self):
        ground_truths, predictions = run_hier_2D_SE_mini_data(self.save_path, n=300, e_a=True, e_s=True)
        return ground_truths, predictions

    def evaluate(self, ground_truths, predictions):
        """
        Evaluate the model.
        """
        # Calculate Normalized Mutual Information (NMI)
        nmi = metrics.normalized_mutual_info_score(ground_truths, predictions)
        print(f"Normalized Mutual Information (NMI): {nmi}")

        # Calculate Adjusted Mutual Information (AMI)
        ami = metrics.adjusted_mutual_info_score(ground_truths, predictions)
        print(f"Adjusted Mutual Information (AMI): {ami}")

        # Calculate Adjusted Rand Index (ARI)
        ari = metrics.adjusted_rand_score(ground_truths, predictions)
        print(f"Adjusted Rand Index (ARI): {ari}")

class Preprocessor:
    def __init__(self, dataset):
        pass

    def get_data_test_df(self, save_path, dataset):
        df = dataset
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        test_set_df_np_path = save_path + 'test_set.npy'

        # 1. 检查 test_set.npy 是否已经存在，如果已存在，检查其是否符合预期
        if os.path.exists(test_set_df_np_path):
            print(f"File {test_set_df_np_path} already exists.")
            # 尝试加载现有的 test_set.npy 文件并检查其内容
            try:
                test_set = np.load(test_set_df_np_path)
                print(f"Loaded test set from {test_set_df_np_path}.")
                print(f"Shape of the array: {test_set.shape}")
                print(f"Data type of the array: {test_set.dtype}")
                print(f"First few rows:\n{test_set[:5]}")  # 打印前 5 行数据
            except Exception as e:
                print(f"Error loading {test_set_df_np_path}: {e}")
                print("The file might be corrupted or empty. Re-generating the file.")
        else:
            print(f"File {test_set_df_np_path} does not exist. Generating the test set...")

            # 2. 加载 test_mask 文件
            try:
                test_mask = torch.load(save_path + 'masks/test_mask.pt').cpu().detach().numpy()
                print("Test mask loaded successfully.")
            except Exception as e:
                print(f"Error loading test_mask.pt: {e}")
                return

            # 3. 检查 test_mask 是否有效，确保掩码的尺寸和内容是合理的
            print(f"Total samples in test mask: {len(test_mask)}")
            print(f"Number of True values in test mask: {np.sum(test_mask)}")
            if np.sum(test_mask) == 0:
                print("Warning: No True values in test mask! No test samples selected.")
                return
            if np.sum(test_mask) == len(test_mask):
                print("Warning: All values in test mask are True! This might not be correct.")
            
            # 4. 根据 test_mask 筛选数据
            test_mask = list(np.where(test_mask == True)[0])  # 获取所有 True 的索引
            test_df = df.iloc[test_mask]
            print(f"Number of samples in the test set: {len(test_df)}")
            if len(test_df) == 0:
                print("Error: The test set is empty after applying the mask.")
                return

            # 5. 将筛选后的 test_df 转换为 NumPy 数组
            test_df_np = test_df.to_numpy()

            # 6. 检查转换后的 test_df 是否为空或维度不匹配
            if test_df_np.shape[0] == 0:
                print("Error: Converted test set is empty.")
                return
            if test_df_np.shape[1] != df.shape[1]:
                print(f"Error: Column mismatch between original data and test set. Original columns: {df.shape[1]}, Test set columns: {test_df_np.shape[1]}")
                return

            # 7. 保存 test_set.npy
            try:
                np.save(test_set_df_np_path, test_df_np)
                print(f"Test set saved successfully to {test_set_df_np_path}.")
            except Exception as e:
                print(f"Error saving {test_set_df_np_path}: {e}")

            # 8. 重新加载并检查保存的文件
            try:
                test_set = np.load(test_set_df_np_path)
                print(f"Loaded saved test set from {test_set_df_np_path}.")
                print(f"Shape of the array: {test_set.shape}")
                print(f"Data type of the array: {test_set.dtype}")
                print(f"First few rows:\n{test_set[:5]}")  # 打印前 5 行数据
            except Exception as e:
                print(f"Error loading {test_set_df_np_path}: {e}")
                print("The file might be corrupted or empty.")
                return

        return

    def get_data_messages_embeddings(self, save_path, language):

        SBERT_embedding_path = f'{save_path}/SBERT_embeddings.pkl'
        if not exists(SBERT_embedding_path):
            test_set_df_np_path = save_path + 'test_set.npy'
            test_df_np = np.load(test_set_df_np_path, allow_pickle=True)
            test_df = pd.DataFrame(data=test_df_np,
                                   columns=['tweet_id', 'text', 'event_id', 'words', 'filtered_words',
                                        'entities', 'user_id', 'created_at', 'urls', 'hashtags', 'user_mentions'
                                        ])
            print("Dataframe loaded.")
            # print('test_df.head(5): \n', test_df.head(5))
            # print('test_df.head(5)[\'text\']: \n', test_df.head(5)['text'])

            processed_text = [preprocess_sentence(s) for s in
                              test_df['text'].values]  # hastags are kept (with '#' removed). RTs are removed.
            print('message text contents preprocessed.')

            # get SBERT embeddings
            embeddings = SBERT_embed(processed_text, language)

            # store SBERT embeddings
            with open(SBERT_embedding_path, 'wb') as fp:
                pickle.dump(embeddings, fp)
            print('SBERT embeddings stored.')
        return


    def split_and_save_masks(self, df, save_dir, train_size=0.7, val_size=0.1, test_size=0.2, random_seed=42):
        """
        Splits the DataFrame into training, validation, and test sets, and saves the indices (masks) as .pt files.
        
        Parameters:
        - df (pd.DataFrame): The DataFrame to be split.
        - train_size (float): Proportion of data to be used for training (default is 70%).
        - val_size (float): Proportion of data to be used for validation (default is 10%).
        - test_size (float): Proportion of data to be used for testing (default is 20%).
        - random_seed (int): Seed for reproducibility (default is 42).
        - save_dir (str): Directory where the masks will be saved.
        
        Returns:
        - None
        """
        

        # 1. Check if the sizes add up to 1
        if train_size + val_size + test_size != 1.0:
            raise ValueError("train_size + val_size + test_size must equal 1.0")
        
        # 2. Ensure the dataframe is not empty
        if df.empty:
            raise ValueError("The input DataFrame is empty.")
        
        print(f"Total samples in DataFrame: {len(df)}")

        # 3. Set random seed for reproducibility
        torch.manual_seed(random_seed)
        
        # 4. Split data into train + temp (for validation and test)
        print("Splitting data into train and temp...")
        train_data, temp_data = train_test_split(df, train_size=train_size, random_state=random_seed)
        
        # 5. Split the temp data into validation and test
        print("Splitting temp data into validation and test...")
        val_data, test_data = train_test_split(temp_data, train_size=val_size / (val_size + test_size), random_state=random_seed)
        
        # 6. Create index masks for each split
        train_mask = train_data.index.tolist()
        val_mask = val_data.index.tolist()
        test_mask = test_data.index.tolist()

        # 7. Create a boolean mask (True for the data points in the respective set, False otherwise)
        # Initialize all as False
        full_train_mask = torch.zeros(len(df), dtype=torch.bool)
        full_val_mask = torch.zeros(len(df), dtype=torch.bool)
        full_test_mask = torch.zeros(len(df), dtype=torch.bool)

        # Mark the corresponding indices as True
        full_train_mask[train_mask] = True
        full_val_mask[val_mask] = True
        full_test_mask[test_mask] = True

        # Print the sizes of each split for verification
        print(f"Training data size: {full_train_mask.sum()} samples")
        print(f"Validation data size: {full_val_mask.sum()} samples")
        print(f"Test data size: {full_test_mask.sum()} samples")

        # 8. Create directory to save masks if it doesn't exist
        os.makedirs(save_dir, exist_ok=True)

        # 9. Define the mask file paths
        mask_paths = {
            'train_mask.pt': full_train_mask,
            'val_mask.pt': full_val_mask,
            'test_mask.pt': full_test_mask
        }

        # 10. Save the masks if the files do not exist
        for filename, mask in mask_paths.items():
            mask_path = os.path.join(save_dir, filename)
            
            if not os.path.exists(mask_path):
                try:
                    # Save the mask as a tensor
                    torch.save(mask, mask_path)
                    print(f"Saved {filename} at {mask_path}")
                except Exception as e:
                    print(f"Error saving {filename} at {mask_path}: {e}")
            else:
                print(f"{filename} already exists at {mask_path}")
            
            # 11. After saving, verify if the file is non-empty
            if os.path.exists(mask_path):
                saved_mask = torch.load(mask_path)
                if saved_mask.numel() == 0:
                    print(f"Warning: {filename} is empty.")
                else:
                    print(f"Successfully saved and verified {filename} with {saved_mask.numel()} elements.")

        print("Mask generation and saving process completed.")

    # Example usage
    # df = pd.DataFrame({...})  # Assuming df is an existing DataFrame
    # split_and_save_masks(df)


    def preprocess_data(self, save_path, dataset, language):
        # load raw data

        print("Data converted to dataframe.")

        save_dir = os.path.join(save_path, 'masks')
        self.split_and_save_masks(dataset, save_dir)
        # get test set df
        self.get_data_test_df(save_path, dataset)

        # get SBERT embeddings
        self.get_data_messages_embeddings(save_path, language)

        return

def get_stable_point(path):
    stable_point_path = path + 'stable_point.pkl'
    if not exists(stable_point_path):
        embeddings_path = path + 'SBERT_embeddings.pkl'
        with open(embeddings_path, 'rb') as f:
            embeddings = pickle.load(f)
        print(f"Loaded embeddings: {embeddings}")
        print(f"Shape of embeddings: {np.shape(embeddings)}")

        first_stable_point, global_stable_point = search_stable_points(embeddings)
        stable_points = {'first': first_stable_point, 'global': global_stable_point}
        with open(stable_point_path, 'wb') as fp:
            pickle.dump(stable_points, fp)
        print('stable points stored.')

    with open(stable_point_path, 'rb') as f:
        stable_points = pickle.load(f)
    print('stable points loaded.')
    return stable_points

def run_hier_2D_SE_mini_data(save_path, n=300, e_a=True, e_s=True):

    # load test_set_df
    test_set_df_np_path = save_path + 'test_set.npy'
    test_df_np = np.load(test_set_df_np_path, allow_pickle=True)
    test_df = pd.DataFrame(data=test_df_np,
                           columns=['tweet_id', 'text', 'event_id', 'words', 'filtered_words',
                                        'entities', 'user_id', 'created_at', 'urls', 'hashtags', 'user_mentions'
                                        ])
    print("Dataframe loaded.")
    all_node_features = [[str(u)] + \
                         [str(each) for each in um] + \
                         [h.lower() for h in hs] + \
                         e \
                         for u, um, hs, e in \
                         zip(test_df['user_id'], test_df['user_mentions'], test_df['hashtags'], test_df['entities'])]

    # load embeddings of the test set messages
    with open(f'{save_path}/SBERT_embeddings.pkl', 'rb') as f:
        embeddings = pickle.load(f)

    stable_points = get_stable_point(save_path)
    default_num_neighbors = stable_points['first']

    global_edges = get_global_edges(all_node_features, embeddings, default_num_neighbors, e_a=e_a, e_s=e_s)
    corr_matrix = np.corrcoef(embeddings)

    np.fill_diagonal(corr_matrix, 0)
    weighted_global_edges = [(edge[0], edge[1], corr_matrix[edge[0] - 1, edge[1] - 1]) for edge in global_edges \
                             if corr_matrix[edge[0] - 1, edge[1] - 1] > 0]  # node encoding starts from 1

    division = hier_2D_SE_mini(weighted_global_edges, len(embeddings), n=n)
    prediction = decode(division)

    labels_true = test_df['event_id'].tolist()
    n_clusters = len(list(set(labels_true)))
    print('n_clusters gt: ', n_clusters)

    return labels_true, prediction




def search_stable_points(embeddings, max_num_neighbors=50):
    corr_matrix = np.corrcoef(embeddings)

    np.fill_diagonal(corr_matrix, 0)
    corr_matrix_sorted_indices = np.argsort(corr_matrix)

    all_1dSEs = []
    seg = None
    for i in range(max_num_neighbors):
        dst_ids = corr_matrix_sorted_indices[:, -(i + 1)]
        knn_edges = [(s + 1, d + 1, corr_matrix[s, d]) \
                     for s, d in enumerate(dst_ids) if
                     corr_matrix[s, d] > 0]  # (s+1, d+1): +1 as node indexing starts from 1 instead of 0
        if i == 0:
            g = nx.Graph()
            g.add_weighted_edges_from(knn_edges)
            seg = SE(g)
            all_1dSEs.append(seg.calc_1dSE())
        else:
            all_1dSEs.append(seg.update_1dSE(all_1dSEs[-1], knn_edges))

    # print('all_1dSEs: ', all_1dSEs)
    stable_indices = []
    for i in range(1, len(all_1dSEs) - 1):
        if all_1dSEs[i] < all_1dSEs[i - 1] and all_1dSEs[i] < all_1dSEs[i + 1]:
            stable_indices.append(i)
    if len(stable_indices) == 0:
        print('No stable points found after checking k = 1 to ', max_num_neighbors)
        return 0, 0
    else:
        stable_SEs = [all_1dSEs[index] for index in stable_indices]
        index = stable_indices[stable_SEs.index(min(stable_SEs))]
        print('stable_indices: ', stable_indices)
        print('stable_SEs: ', stable_SEs)
        print('First stable point: k = ', stable_indices[0] + 1, ', correspoding 1dSE: ',
              stable_SEs[0])  # n_neighbors should be index + 1
        print('Global stable point within the searching range: k = ', index + 1, \
              ', correspoding 1dSE: ', all_1dSEs[index])  # n_neighbors should be index + 1
    return stable_indices[0] + 1, index + 1  # first stable point, global stable point

def get_graph_edges(attributes):
    attr_nodes_dict = {}
    for i, l in enumerate(attributes):
        for attr in l:
            if attr not in attr_nodes_dict:
                attr_nodes_dict[attr] = [i + 1]  # node indexing starts from 1
            else:
                attr_nodes_dict[attr].append(i + 1)

    for attr in attr_nodes_dict.keys():
        attr_nodes_dict[attr].sort()

    graph_edges = []
    for l in attr_nodes_dict.values():
        graph_edges += list(combinations(l, 2))
    return list(set(graph_edges))

def get_knn_edges(embeddings, default_num_neighbors):
    corr_matrix = np.corrcoef(embeddings)
    np.fill_diagonal(corr_matrix, 0)
    corr_matrix_sorted_indices = np.argsort(corr_matrix)
    knn_edges = []
    for i in range(default_num_neighbors):
        dst_ids = corr_matrix_sorted_indices[:, -(i + 1)]
        knn_edges += [(s + 1, d + 1) if s < d else (d + 1, s + 1) \
                      for s, d in enumerate(dst_ids) if
                      corr_matrix[s, d] > 0]  # (s+1, d+1): +1 as node indexing starts from 1 instead of 0
    return list(set(knn_edges))

def get_global_edges(attributes, embeddings, default_num_neighbors, e_a=True, e_s=True):
    graph_edges, knn_edges = [], []
    if e_a == True:
        graph_edges = get_graph_edges(attributes)
    if e_s == True:
        knn_edges = get_knn_edges(embeddings, default_num_neighbors)
    return list(set(knn_edges + graph_edges))

def get_subgraphs_edges(clusters, graph_splits, weighted_global_edges):
    '''
    get the edges of each subgraph

    clusters: a list containing the current clusters, each cluster is a list of nodes of the original graph
    graph_splits: a list of (start_index, end_index) pairs, each (start_index, end_index) pair indicates a subset of clusters, 
        which will serve as the nodes of a new subgraph
    weighted_global_edges: a list of (start node, end node, edge weight) tuples, each tuple is an edge in the original graph

    return: all_subgraphs_edges: a list containing the edges of all subgraphs
    '''
    all_subgraphs_edges = []
    for split in graph_splits:
        subgraph_clusters = clusters[split[0]:split[1]]
        subgraph_nodes = list(chain(*subgraph_clusters))
        subgraph_edges = [edge for edge in weighted_global_edges if
                          edge[0] in subgraph_nodes and edge[1] in subgraph_nodes]
        all_subgraphs_edges.append(subgraph_edges)
    return all_subgraphs_edges

def hier_2D_SE_mini(weighted_global_edges, n_messages, n=100):
    '''
    hierarchical 2D SE minimization
    '''
    ite = 1
    # initially, each node (message) is in its own cluster
    # node encoding starts from 1
    clusters = [[i + 1] for i in range(n_messages)]
    while True:
        print('\n=========Iteration ', str(ite), '=========')
        n_clusters = len(clusters)
        graph_splits = [(s, min(s + n, n_clusters)) for s in range(0, n_clusters, n)]  # [s, e)
        all_subgraphs_edges = get_subgraphs_edges(clusters, graph_splits, weighted_global_edges)
        last_clusters = clusters
        clusters = []
        print('111111111111111111111111111Number of subgraphs: ', len(graph_splits))
        for i, subgraph_edges in enumerate(all_subgraphs_edges):
            print('\tSubgraph ', str(i + 1))
            g = nx.Graph()
            g.add_weighted_edges_from(subgraph_edges)
            seg = SE(g)
            seg.division = {j: cluster for j, cluster in
                            enumerate(last_clusters[graph_splits[i][0]:graph_splits[i][1]])}
            seg.add_isolates()
            for k in seg.division.keys():
                for node in seg.division[k]:
                    seg.graph.nodes[node]['comm'] = k
            seg.update_struc_data()
            seg.update_struc_data_2d()
            seg.update_division_MinSE()

            clusters += list(seg.division.values())
        if len(graph_splits) == 1:
            break
        if clusters == last_clusters:
            n *= 2
    return clusters

# SE
class SE:
    def __init__(self, graph: nx.Graph):
        self.graph = graph.copy()
        self.vol = self.get_vol()
        self.division = {}  # {comm1: [node11, node12, ...], comm2: [node21, node22, ...], ...}
        self.struc_data = {}  # {comm1: [vol1, cut1, community_node_SE, leaf_nodes_SE], comm2:[vol2, cut2, community_node_SE, leaf_nodes_SE]，... }
        self.struc_data_2d = {}  # {comm1: {comm2: [vol_after_merge, cut_after_merge, comm_node_SE_after_merge, leaf_nodes_SE_after_merge], comm3: [], ...}, ...}

    def get_vol(self):
        '''
        get the volume of the graph
        '''
        return cuts.volume(self.graph, self.graph.nodes, weight='weight')

    def calc_1dSE(self):
        '''
        get the 1D SE of the graph
        '''
        SE = 0
        for n in self.graph.nodes:
            d = cuts.volume(self.graph, [n], weight='weight')
            SE += - (d / self.vol) * math.log2(d / self.vol)
        return SE

    def update_1dSE(self, original_1dSE, new_edges):
        '''
        get the updated 1D SE after new edges are inserted into the graph
        '''

        affected_nodes = []
        for edge in new_edges:
            affected_nodes += [edge[0], edge[1]]
        affected_nodes = set(affected_nodes)

        original_vol = self.vol
        original_degree_dict = {node: 0 for node in affected_nodes}
        for node in affected_nodes.intersection(set(self.graph.nodes)):
            original_degree_dict[node] = self.graph.degree(node, weight='weight')

        # insert new edges into the graph
        self.graph.add_weighted_edges_from(new_edges)

        self.vol = self.get_vol()
        updated_vol = self.vol
        updated_degree_dict = {}
        for node in affected_nodes:
            updated_degree_dict[node] = self.graph.degree(node, weight='weight')

        updated_1dSE = (original_vol / updated_vol) * (original_1dSE - math.log2(original_vol / updated_vol))
        for node in affected_nodes:
            d_original = original_degree_dict[node]
            d_updated = updated_degree_dict[node]
            if d_original != d_updated:
                if d_original != 0:
                    updated_1dSE += (d_original / updated_vol) * math.log2(d_original / updated_vol)
                updated_1dSE -= (d_updated / updated_vol) * math.log2(d_updated / updated_vol)

        return updated_1dSE

    def get_cut(self, comm):
        '''
        get the sum of the degrees of the cut edges of community comm
        '''
        return cuts.cut_size(self.graph, comm, weight='weight')

    def get_volume(self, comm):
        '''
        get the volume of community comm
        '''
        return cuts.volume(self.graph, comm, weight='weight')

    def calc_2dSE(self):
        '''
        get the 2D SE of the graph
        '''
        SE = 0
        for comm in self.division.values():
            g = self.get_cut(comm)
            v = self.get_volume(comm)
            SE += - (g / self.vol) * math.log2(v / self.vol)
            for node in comm:
                d = self.graph.degree(node, weight='weight')
                SE += - (d / self.vol) * math.log2(d / v)
        return SE

    def show_division(self):
        print(self.division)

    def show_struc_data(self):
        print(self.struc_data)

    def show_struc_data_2d(self):
        print(self.struc_data_2d)

    def print_graph(self):
        fig, ax = plt.subplots()
        nx.draw(self.graph, ax=ax, with_labels=True)
        plt.show()

    def update_struc_data(self):
        '''
        calculate the volume, cut, communitiy mode SE, and leaf nodes SE of each cummunity, 
        then store them into self.struc_data
        '''
        self.struc_data = {}  # {comm1: [vol1, cut1, community_node_SE, leaf_nodes_SE], comm2:[vol2, cut2, community_node_SE, leaf_nodes_SE]，... }
        for vname in self.division.keys():
            comm = self.division[vname]
            volume = self.get_volume(comm)
            cut = self.get_cut(comm)
            if volume == 0:
                vSE = 0
            else:
                vSE = - (cut / self.vol) * math.log2(volume / self.vol)
            vnodeSE = 0
            for node in comm:
                d = self.graph.degree(node, weight='weight')
                if d != 0:
                    vnodeSE -= (d / self.vol) * math.log2(d / volume)
            self.struc_data[vname] = [volume, cut, vSE, vnodeSE]

    def update_struc_data_2d(self):
        '''
        calculate the volume, cut, communitiy mode SE, and leaf nodes SE after merging each pair of cummunities, 
        then store them into self.struc_data_2d
        '''
        self.struc_data_2d = {}  # {(comm1, comm2): [vol_after_merge, cut_after_merge, comm_node_SE_after_merge, leaf_nodes_SE_after_merge], (comm1, comm3): [], ...}
        comm_num = len(self.division)
        for i in range(comm_num):
            for j in range(i + 1, comm_num):
                v1 = list(self.division.keys())[i]
                v2 = list(self.division.keys())[j]
                if v1 < v2:
                    k = (v1, v2)
                else:
                    k = (v2, v1)

                comm_merged = self.division[v1] + self.division[v2]
                gm = self.get_cut(comm_merged)
                vm = self.struc_data[v1][0] + self.struc_data[v2][0]
                if self.struc_data[v1][0] == 0 or self.struc_data[v2][0] == 0:
                    vmSE = self.struc_data[v1][2] + self.struc_data[v2][2]
                    vmnodeSE = self.struc_data[v1][3] + self.struc_data[v2][3]
                else:
                    vmSE = - (gm / self.vol) * math.log2(vm / self.vol)
                    vmnodeSE = self.struc_data[v1][3] - (self.struc_data[v1][0] / self.vol) * math.log2(
                        self.struc_data[v1][0] / vm) + \
                               self.struc_data[v2][3] - (self.struc_data[v2][0] / self.vol) * math.log2(
                        self.struc_data[v2][0] / vm)
                self.struc_data_2d[k] = [vm, gm, vmSE, vmnodeSE]

    def init_division(self):
        '''
        initialize self.division such that each node assigned to its own community
        '''
        self.division = {}
        for node in self.graph.nodes:
            new_comm = node
            self.division[new_comm] = [node]
            self.graph.nodes[node]['comm'] = new_comm

    def add_isolates(self):
        '''
        add any isolated nodes into graph
        '''
        all_nodes = list(chain(*list(self.division.values())))
        all_nodes.sort()
        edge_nodes = list(self.graph.nodes)
        edge_nodes.sort()
        if all_nodes != edge_nodes:
            for node in set(all_nodes) - set(edge_nodes):
                self.graph.add_node(node)

    def update_division_MinSE(self):
        '''
        greedily update the encoding tree to minimize 2D SE
        '''

        def Mg_operator(v1, v2):
            '''
            MERGE operator. It calculates the delta SE caused by mergeing communities v1 and v2, 
            without actually merging them, i.e., the encoding tree won't be changed
            '''
            v1SE = self.struc_data[v1][2]
            v1nodeSE = self.struc_data[v1][3]

            v2SE = self.struc_data[v2][2]
            v2nodeSE = self.struc_data[v2][3]

            if v1 < v2:
                k = (v1, v2)
            else:
                k = (v2, v1)
            vm, gm, vmSE, vmnodeSE = self.struc_data_2d[k]
            delta_SE = vmSE + vmnodeSE - (v1SE + v1nodeSE + v2SE + v2nodeSE)
            return delta_SE

        # continue merging any two communities that can cause the largest decrease in SE, 
        # until the SE can't be further reduced
        while True:
            comm_num = len(self.division)
            delta_SE = 99999
            vm1 = None
            vm2 = None
            for i in range(comm_num):
                for j in range(i + 1, comm_num):
                    v1 = list(self.division.keys())[i]
                    v2 = list(self.division.keys())[j]
                    new_delta_SE = Mg_operator(v1, v2)
                    if new_delta_SE < delta_SE:
                        delta_SE = new_delta_SE
                        vm1 = v1
                        vm2 = v2

            if delta_SE < 0:
                # Merge v2 into v1, and update the encoding tree accordingly
                for node in self.division[vm2]:
                    self.graph.nodes[node]['comm'] = vm1
                self.division[vm1] += self.division[vm2]
                self.division.pop(vm2)

                volume = self.struc_data[vm1][0] + self.struc_data[vm2][0]
                cut = self.get_cut(self.division[vm1])
                vmSE = - (cut / self.vol) * math.log2(volume / self.vol)
                vmnodeSE = self.struc_data[vm1][3] - (self.struc_data[vm1][0] / self.vol) * math.log2(
                    self.struc_data[vm1][0] / volume) + \
                           self.struc_data[vm2][3] - (self.struc_data[vm2][0] / self.vol) * math.log2(
                    self.struc_data[vm2][0] / volume)
                self.struc_data[vm1] = [volume, cut, vmSE, vmnodeSE]
                self.struc_data.pop(vm2)

                struc_data_2d_new = {}
                for k in self.struc_data_2d.keys():
                    if k[0] == vm2 or k[1] == vm2:
                        continue
                    elif k[0] == vm1 or k[1] == vm1:
                        v1 = k[0]
                        v2 = k[1]
                        comm_merged = self.division[v1] + self.division[v2]
                        gm = self.get_cut(comm_merged)
                        vm = self.struc_data[v1][0] + self.struc_data[v2][0]
                        if self.struc_data[v1][0] == 0 or self.struc_data[v2][0] == 0:
                            vmSE = self.struc_data[v1][2] + self.struc_data[v2][2]
                            vmnodeSE = self.struc_data[v1][3] + self.struc_data[v2][3]
                        else:
                            vmSE = - (gm / self.vol) * math.log2(vm / self.vol)
                            vmnodeSE = self.struc_data[v1][3] - (self.struc_data[v1][0] / self.vol) * math.log2(
                                self.struc_data[v1][0] / vm) + \
                                       self.struc_data[v2][3] - (self.struc_data[v2][0] / self.vol) * math.log2(
                                self.struc_data[v2][0] / vm)
                        struc_data_2d_new[k] = [vm, gm, vmSE, vmnodeSE]
                    else:
                        struc_data_2d_new[k] = self.struc_data_2d[k]
                self.struc_data_2d = struc_data_2d_new
            else:
                break

def vanilla_2D_SE_mini(weighted_edges):
    """
    vanilla (greedy) 2D SE minimization
    """
    g = nx.Graph()
    g.add_weighted_edges_from(weighted_edges)

    seg = SE(g)
    seg.init_division()
    # seg.show_division()
    SE1D = seg.calc_1dSE()

    seg.update_struc_data()
    # seg.show_struc_data()
    seg.update_struc_data_2d()
    # seg.show_struc_data_2d()
    initial_SE2D = seg.calc_2dSE()

    seg.update_division_MinSE()
    communities = seg.division
    minimized_SE2D = seg.calc_2dSE()

    return SE1D, initial_SE2D, minimized_SE2D, communities

def test_vanilla_2D_SE_mini():
    weighted_edges = [(1, 2, 2), (1, 3, 4)]

    g = nx.Graph()
    g.add_weighted_edges_from(weighted_edges)
    A = nx.adjacency_matrix(g).todense()
    print('adjacency matrix: \n', A)
    print('g.nodes: ', g.nodes)
    print('g.edges: ', g.edges)
    print('degrees of nodes: ', list(g.degree(g.nodes, weight='weight')))

    SE1D, initial_SE2D, minimized_SE2D, communities = vanilla_2D_SE_mini(weighted_edges)
    print('\n1D SE of the graph: ', SE1D)
    print('initial 2D SE of the graph: ', initial_SE2D)
    print('the minimum 2D SE of the graph: ', minimized_SE2D)
    print('communities detected: ', communities)
    return

# utils

def replaceAtUser(text):
    """ Replaces "@user" with "" """
    text = re.sub('@[^\s]+|RT @[^\s]+', '', text)
    return text


def removeUnicode(text):
    """ Removes unicode strings like "\u002c" and "x96" """
    text = re.sub(r'(\\u[0-9A-Fa-f]+)', r'', text)
    text = re.sub(r'[^\x00-\x7f]', r'', text)
    return text


def replaceURL(text):
    """ Replaces url address with "url" """
    text = re.sub('((www\.[^\s]+)|(https?://[^\s]+))', 'url', text)
    text = re.sub(r'#([^\s]+)', r'\1', text)
    return text


def replaceMultiExclamationMark(text):
    """ Replaces repetitions of exlamation marks """
    text = re.sub(r"(\!)\1+", '!', text)
    return text


def replaceMultiQuestionMark(text):
    """ Replaces repetitions of question marks """
    text = re.sub(r"(\?)\1+", '?', text)
    return text


def removeEmoticons(text):
    """ Removes emoticons from text """
    text = re.sub(
        ':\)|;\)|:-\)|\(-:|:-D|=D|:P|xD|X-p|\^\^|:-*|\^\.\^|\^\-\^|\^\_\^|\,-\)|\)-:|:\'\(|:\(|:-\(|:\S|T\.T|\.\_\.|:<|:-\S|:-<|\*\-\*|:O|=O|=\-O|O\.o|XO|O\_O|:-\@|=/|:/|X\-\(|>\.<|>=\(|D:',
        '', text)
    return text


def removeNewLines(text):
    text = re.sub('\n', '', text)
    return text


def preprocess_sentence(s):
    return removeNewLines(replaceAtUser(
        removeEmoticons(replaceMultiQuestionMark(replaceMultiExclamationMark(removeUnicode(replaceURL(s)))))))


def preprocess_french_sentence(s):
    return removeNewLines(
        replaceAtUser(removeEmoticons(replaceMultiQuestionMark(replaceMultiExclamationMark(replaceURL(s))))))


def SBERT_embed(s_list, language):
    '''
    Use Sentence-BERT to embed sentences.
    s_list: a list of sentences/ tokens to be embedded.
    language: the language of the sentences ('English', 'French', 'Arabic').
    output: the embeddings of the sentences/ tokens.
    '''
    # Model paths or names for each language
    model_map = {
        'English': '../model/model_needed/all-MiniLM-L6-v2',
        'French': '../model/model_needed/distiluse-base-multilingual-cased-v1',
        'Arabic': '../model/model_needed/paraphrase-multilingual-mpnet-base-v2'
    }

    # Default model for Hugging Face
    hf_model_map = {
        'English': 'sentence-transformers/all-MiniLM-L6-v2',
        'French': 'sentence-transformers/distiluse-base-multilingual-cased-v1',
        'Arabic': 'sentence-transformers/paraphrase-multilingual-mpnet-base-v2'
    }

    # Print language and model being used
    print(f"Embedding sentences in language: {language}")
    
    # Determine model path
    model_path = model_map.get(language)
    if not model_path:
        raise ValueError(f"Unsupported language: {language}. Supported languages are: {', '.join(model_map.keys())}")

    print(f"Using model: {model_path}")

    # Load the model, downloading if necessary
    try:
        model = SentenceTransformer(model_path)
        print(f"Successfully loaded model from local path: {model_path}")
    except Exception as e:
        print(f"Model {model_path} not found locally. Attempting to download from Hugging Face...")
        model = SentenceTransformer(hf_model_map[language])
        print(f"Model downloaded from Hugging Face: {hf_model_map[language]}")

    # Compute embeddings
    embeddings = model.encode(s_list, convert_to_tensor=True, normalize_embeddings=True)
    print(f"Computed embeddings for {len(s_list)} sentences/tokens.")
    
    return embeddings.cpu()



def evaluate(labels_true, labels_pred):
    nmi = normalized_mutual_info_score(labels_true, labels_pred)
    ami = adjusted_mutual_info_score(labels_true, labels_pred)
    ari = adjusted_rand_score(labels_true, labels_pred)
    return nmi, ami, ari


def decode(division):
    if type(division) is dict:
        prediction_dict = {m: event for event, messages in division.items() for m in messages}
    elif type(division) is list:
        prediction_dict = {m: event for event, messages in enumerate(division) for m in messages}
    prediction_dict_sorted = dict(sorted(prediction_dict.items()))
    return list(prediction_dict_sorted.values())


if __name__ == "__main__":

    #Data = DatasetLoader("MAVEN")
    Data = DatasetLoader_mini("Event2012")

    dataset = Data.load_data()
    language = Data.get_dataset_language()
    hisevent = HISEvent(dataset, language)
    hisevent.preprocess()
    predictions, ground_truths = hisevent.detection()
    hisevent.evaluate(predictions, ground_truths)






import pytest
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import numpy as np
import pandas as pd
from sklearn.metrics.cluster import normalized_mutual_info_score, adjusted_mutual_info_score, adjusted_rand_score
from sentence_transformers import SentenceTransformer
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dataset.dataloader import DatasetLoader
from detector.HISEvent import *
# Mock data for testing
@pytest.fixture
def mock_dataset():
    data = {
        'filtered_words': [['word1', 'word2'], ['word3', 'word4'], ['word5', 'word6']],
        'event_id': [1, 2, 1]
    }
    return pd.DataFrame(data)

@pytest.fixture
def hisevent_instance():
    return HISEvent()

@pytest.fixture
def preprocessor_instance():
    return Preprocessor()

def test_preprocess(hisevent_instance):
    hisevent_instance.preprocess()
    # Add assertions to check if preprocessing was successful
    assert True

def test_detection(hisevent_instance):
    hisevent_instance.detection()
    # Add assertions to check if detection was successful
    assert True
'''
def test_get_stable_point(tmpdir):
    path = tmpdir.mkdir("test_dir")
    embeddings = np.random.rand(10, 100)
    embeddings_path = os.path.join(path, 'SBERT_embeddings.pkl')
    with open(embeddings_path, 'wb') as f:
        pickle.dump(embeddings, f)
    
    # 确保文件存在
    assert os.path.exists(embeddings_path)
    
    stable_points = get_stable_point(str(path))
    assert isinstance(stable_points, dict)
    assert 'first' in stable_points
    assert 'global' in stable_points
'''
def test_run_hier_2D_SE_mini_Event2012_open_set(tmpdir):
    save_path = str(tmpdir.mkdir("test_dir"))
    embeddings = np.random.rand(10, 100)
    df_np = np.array([[i, i % 3, f'tweet_{i}', f'user_{i}', '2023-01-01'] for i in range(10)])
    df = pd.DataFrame(data=df_np, columns=["original_index", "event_id", "text", "user_id", "created_at"])
    block_path = os.path.join(save_path, '1')
    os.makedirs(block_path, exist_ok=True)
    np.save(os.path.join(block_path, '1.npy'), df_np)
    embeddings_path = os.path.join(block_path, 'SBERT_embeddings.pkl')
    with open(embeddings_path, 'wb') as f:
        pickle.dump(embeddings, f)
    
    # 确保文件存在
    assert os.path.exists(embeddings_path)
    
    run_hier_2D_SE_mini_Event2012_open_set(n=10, e_a=True, e_s=True, test_with_one_block=True)
    # Add assertions to check if the function ran successfully
    assert True

def test_evaluate():
    labels_true = [1, 2, 1]
    labels_pred = [0, 1, 0]
    nmi, ami, ari = evaluate(labels_true, labels_pred)
    assert isinstance(nmi, float)
    assert isinstance(ami, float)
    assert isinstance(ari, float)

def test_decode():
    division = [[1, 2], [3, 4]]
    prediction = decode(division)
    assert isinstance(prediction, list)
    assert len(prediction) == 4

# Additional tests can be added as needed


import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import pytest
import numpy as np
import pandas as pd
import torch
from transformers import BertTokenizer, BertModel
from sklearn import metrics
from dataset.dataloader import DatasetLoader
from detector.BERT import *

# 创建一个示例数据集
@pytest.fixture
def sample_dataset():
    data = {
        'event_id': [1, 2, 1, 2, 3, 1, 2, 3, 1, 2],
        'filtered_words': [
            ['hello', 'world'], ['goodbye', 'world'], ['hello', 'world'], ['goodbye', 'world'], ['new', 'event'],
            ['hello', 'world'], ['goodbye', 'world'], ['new', 'event'], ['hello', 'world'], ['goodbye', 'world']
        ]
    }
    return pd.DataFrame(data)

# 测试数据预处理
def test_preprocess(sample_dataset):
    bert = BERT(sample_dataset)
    df = bert.preprocess()
    assert 'processed_text' in df.columns
    assert df['processed_text'].apply(lambda x: isinstance(x, str)).all()

# 测试获取 BERT 嵌入
def test_get_bert_embeddings(sample_dataset):
    bert = BERT(sample_dataset)
    bert.preprocess()
    text = 'hello world'
    embedding = bert.get_bert_embeddings(text)
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (768,)

# 测试检测
def test_detection(sample_dataset):
    bert = BERT(sample_dataset)
    bert.preprocess()
    ground_truths, predictions = bert.detection()
    assert len(ground_truths) == len(predictions)

# 测试评估
def test_evaluate(sample_dataset):
    bert = BERT(sample_dataset)
    bert.preprocess()
    ground_truths, predictions = bert.detection()
    ari, ami, nmi = bert.evaluate(ground_truths, predictions)
    assert isinstance(ari, float)
    assert isinstance(ami, float)
    assert isinstance(nmi, float)

if __name__ == "__main__":
    pytest.main()